"""Implementation of velph-init."""

from __future__ import annotations

import copy
import dataclasses
import io
import os
import pathlib
from collections.abc import Sequence
from typing import Literal, Optional, Union

try:
    from spglib import SpglibDataset
except ImportError:
    from types import SimpleNamespace as SpglibDataset

import click
import numpy as np
import tomli
import tomli_w
from phono3py.phonon.grid import GridMatrix
from phonopy.interface.calculator import read_crystal_structure
from phonopy.interface.vasp import get_vasp_structure_lines
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import (
    estimate_supercell_matrix,
    get_supercell,
    shape_supercell_matrix,
)

from phelel.velph.cli.utils import (
    CellChoice,
    DefaultCellChoices,
    PrimitiveCellChoice,
    VelphFilePaths,
    VelphInitParams,
    generate_standardized_cells,
    get_primitive_cell,
    get_reduced_cell,
    get_symmetry_dataset,
)
from phelel.velph.templates import default_template_dict
from phelel.velph.utils.vasp import CutoffToFFTMesh, VaspIncar
from phelel.version import __version__


def run_init(
    cmd_init_options: dict, vfp: VelphFilePaths, phelel_dir_name: str = "phelel"
) -> Optional[list[str]]:
    """Run velph-init.

    Main part of init processes are implemented in the function _run_init.

    Preference order of configurations is as follows.

    1. Detailed configurations in [phelel] and [vasp] in velph-template.
    2. Command line options.
    3. [init.options] (alternative of command-line-options) in velph-template.

    Parameters
    ----------
    cmd_options : dict
        Command line options.
    vfp : VelphFilePaths
        Input and output file names required for velph init. Default path to
        scheduler-toml-template file is defined in VelphFilePaths.
    phelel_dir_name : str, optional
        Directory name for [vasp.{phelel_dir_name}]. The default is "phelel",
        which used to be "supercell". This parameter exists for backward
        compatibility.

    Returns
    -------
    list[str, ...]
        velph-toml lines.

    """
    #
    # Parse files.
    #
    input_cell, _ = read_crystal_structure(vfp.cell_filepath, interface_mode="vasp")
    click.echo(f'Read crystal structure file "{vfp.cell_filepath}".')

    return _run_init(
        input_cell,
        cmd_init_options,
        velph_template_fp=vfp.velph_template_filepath,
        template_toml_filepath=vfp.velph_template_filepath,
        phelel_dir_name=phelel_dir_name,
    )


def _run_init(
    input_cell: PhonopyAtoms,
    cmd_init_options: dict,
    velph_template_fp: Optional[Union[str, bytes, os.PathLike, io.IOBase]] = None,
    template_toml_filepath: Optional[Union[str, bytes, os.PathLike]] = None,
    phelel_dir_name: str = "phelel",
) -> Optional[list[str]]:
    """Run init process and return velph-toml lines.

    Parameters
    ----------
    input_cell : PhonopyAtoms
        Input crystal structure.
    cmd_init_options : dict
        Parameters provided by velph-init command options.
    velph_template_fp : str, bytes, os.PathLike, io.IOBase, or None
        velph toml template path. The parameter in str, bytes, or os.PathLike
        represents file name.
    scheduler_template_fp : str, bytes, os.PathLike, io.IOBase, or None
        Scheduler toml template path. The parameter in str, bytes, or os.PathLike
        represents file name.
    template_toml_filepath : str, bytes, os.PathLike
        File name of velph-toml-template.
    phelel_dir_name : str, optional
        Directory name for [vasp.{phelel_dir_name}]. The default is "phelel",
        which used to be "supercell". This parameter exists for backward
        compatibility.

    """
    #
    # Prepare velph configurations: default + template files
    #
    try:
        velph_template_dict = _parse_velph_template(velph_template_fp)
    except tomli.TOMLDecodeError as e:
        click.echo(f'Error in reading "{velph_template_fp}": {e}', err=True)
        return None

    #
    # Collect velph-init command line options.
    #
    template_init_params = _get_template_init_params(velph_template_dict)
    vip = _collect_init_params(
        cmd_init_options, template_init_params, template_toml_filepath
    )

    #
    # Set magnetic moments
    #
    if vip.magmom is not None:
        magmom_vals = VaspIncar().expand(vip.magmom.strip())
        input_cell.magnetic_moments = magmom_vals

    #
    # Define cells and find crystal symmetry.
    #
    unitcell, primitive, sym_dataset = _get_cells(
        input_cell,
        vip.tolerance,
        vip.symmetrize_cell,
        vip.find_primitive,
        vip.primitive_cell_choice,
    )

    #
    # Parse velph configurations.
    #
    velph_dict = _get_velph_dict(velph_template_dict)

    #
    # Determine cell choices for calculations such as nac, relax, etc.
    #
    cell_choices = _determine_cell_choices(vip, velph_dict)

    #
    # Create velph-toml lines.
    #
    toml_lines = _get_toml_lines(
        velph_dict,
        vip,
        unitcell,
        primitive,
        cell_choices,
        sym_dataset,
        phelel_dir_name=phelel_dir_name,
    )

    return toml_lines


def _get_supercell_dimension(
    velph_dict_calc_type: dict,
    max_num_atoms: Optional[int],
    sym_dataset: SpglibDataset,
    find_primitive: bool,
) -> Optional[np.ndarray]:
    """Return supercell dimension.

    This function is used to determine supercell dimension for velph-init.

    Parameters
    ----------
    velph_dict_calc_type : dict
        velph-toml data of calc_type.
    max_num_atoms : int, optional
        Maximum number of atoms in the supercell. Supercell is constructed
        to preserve the point group of the lattice.
    sym_dataset : SpglibDataset
        Symmetry dataset of the input cell.
    find_primitive : bool
        When True, supercell is constructured

    """
    if max_num_atoms is not None:
        if find_primitive is False:
            _max_num_atoms = max_num_atoms * np.rint(
                1.0 / np.linalg.det(sym_dataset.transformation_matrix)
            ).astype(int)
        else:
            _max_num_atoms = max_num_atoms
        supercell_dimension = shape_supercell_matrix(
            estimate_supercell_matrix(sym_dataset, max_num_atoms=_max_num_atoms)
        )
    else:
        try:
            supercell_dimension = shape_supercell_matrix(
                velph_dict_calc_type["supercell_dimension"]
            )
        except KeyError:
            return None

    return supercell_dimension


def _determine_cell_choices(vip: VelphInitParams, velph_dict: dict) -> dict:
    """Determine cell choices for calculations such as nac, relax, etc.

    Cell choices are collected from VelphInitParams. When unspecified in
    VelphInitParams, [vasp.*.cell] in velph dict are examined.

    """
    cell_choices = dataclasses.asdict(DefaultCellChoices())
    for key in ("nac", "relax"):
        value = getattr(vip, f"cell_for_{key}")
        if value is CellChoice.UNSPECIFIED:
            if (
                "vasp" in velph_dict
                and key in velph_dict["vasp"]
                and "cell" in velph_dict["vasp"][key]
            ):
                for _cell_choice in CellChoice:
                    if _cell_choice is CellChoice.UNSPECIFIED:
                        continue
                    if _cell_choice.value in velph_dict["vasp"][key]["cell"]:
                        cell_choices[key] = _cell_choice
        else:
            cell_choices[key] = value
        assert cell_choices[key] in CellChoice
        assert cell_choices[key] is not CellChoice.UNSPECIFIED
    return cell_choices


def _get_template_init_params(velph_template_dict: Optional[dict]) -> dict:
    """Collect init params in [init.options] in velph-toml-template file."""
    template_init_params = {}
    if velph_template_dict:
        if "init" in velph_template_dict and "options" in velph_template_dict["init"]:
            vip_keys = velph_template_dict["init"]["options"].keys()
            for field in dataclasses.fields(VelphInitParams):
                key = field.name
                if key in vip_keys:
                    template_init_params[key] = velph_template_dict["init"]["options"][
                        key
                    ]
    return template_init_params


def _collect_init_params(
    cmd_init_options: dict,
    template_init_params: dict,
    template_toml_filepath: Optional[pathlib.Path],
) -> VelphInitParams:
    """Merge init params defined different places.

    Init parameters were collected in the following order. For the same
    parameters, the latters override the formers.
    1. Defalut VelphInitParams
    2. template_dict["init"]["options"]
    3. Command line options

    Returns
    -------
    VelphInitParams
        Init parameters.

    """
    # Set default parameters.
    vip_dict = {}

    # Set parameters specified in velph-toml-template file.
    for key, value in template_init_params.items():
        if key in ("cell_for_nac", "cell_for_relax"):
            for cell_choice in CellChoice:
                if value.lower() == cell_choice.value:
                    vip_dict[key] = cell_choice
        elif key == "primitive_cell_choice":
            for primitive_cell_choice in PrimitiveCellChoice:
                if value.lower() == primitive_cell_choice.value:
                    vip_dict[key] = primitive_cell_choice
        else:
            vip_dict[key] = value

    # Show parameters specified in velph-toml-template file.
    if template_toml_filepath:
        if len(template_init_params) == 1:
            click.echo("Following option was found in " f'"{template_toml_filepath}":')
        elif len(template_init_params) > 1:
            click.echo(
                "Following options were found in " f'"{template_toml_filepath}":'
            )
    if template_init_params:
        click.echo(
            "\n".join(
                [f'  {key} = "{value}"' for key, value in template_init_params.items()]
            )
        )

    # Collect parameters specified by command-line options.
    # Filling by None for all keys in VelphInitParams is for test mimicing
    # the behaviour of command-line-options that would already fill so.
    cmd_params = {}
    num_active_cmd_params = 0
    for key in dataclasses.asdict(VelphInitParams()):
        if key in cmd_init_options:
            value = cmd_init_options[key]
            if value is not None:
                num_active_cmd_params += 1
            cmd_params[key] = value
        else:
            cmd_params[key] = None

    # Show parameters specified by command-line options.
    if num_active_cmd_params == 1:
        click.echo("Following option was given by command-line options:")
    elif num_active_cmd_params > 1:
        click.echo("Following options were given by command-line options:")

    for key, value in cmd_params.items():
        if value is None:
            continue
        if key in ("cell_for_nac", "cell_for_relax"):
            for cell_choice in CellChoice:
                if cell_choice == CellChoice.UNSPECIFIED:
                    continue
                if value.lower() == cell_choice.value:
                    cmd_params[key] = cell_choice
                    click.echo(f'  {key} = "{value.lower()}"')
        elif key == "primitive_cell_choice":
            for primitive_cell_choice in PrimitiveCellChoice:
                if value.lower() == primitive_cell_choice.value:
                    cmd_params[key] = primitive_cell_choice
        else:
            click.echo(f'  {key} = "{value}"')

    shared_params = [
        key_c == key_t for key_t in template_init_params for key_c in cmd_params
    ]
    if template_toml_filepath:
        if sum(shared_params) == 1:
            click.echo(
                "The command option was prefered to [init.options] in "
                f'"{template_toml_filepath}".'
            )
        if sum(shared_params) > 1:
            click.echo(
                "The command options were prefered to [init.options] in "
                f'"{template_toml_filepath}".'
            )

    # Set parameters specified by command-line options. vip_dict is already
    # filled by [init.options]. But command-line-options have higher preference.
    for key, value in cmd_params.items():
        if value is not None:
            vip_dict[key] = value

    vip = VelphInitParams(**vip_dict)
    return vip


def _get_cells(
    input_cell: PhonopyAtoms,
    tolerance: float,
    symmetrize_cell: bool,
    find_primitive: bool,
    primitive_cell_choice: PrimitiveCellChoice,
) -> tuple[PhonopyAtoms, PhonopyAtoms, dict]:
    """Return unit cell, primitive cell, and symmetry dataset.

    This function is complicated due to complicated requests.

    The unit cell and primitive cell returned can change depending on
    `symmetrize_cell`, `find_primitive` and `primitive_cell_choice`. When
    `primitive_cell_choice==PrimitiveCellChoice.REDUCED`, the primitive cell
    is finally redued.

    * `symmetrize_cell=True` and `find_primitive=True`.

    The unit cell and primitive cell are those standardized by spglib.

    * `symmetrize_cell=False` and `find_primitive=True`.

    The unit cell is the input cell. If the input cell is a primitive cell, it
    is used as the primitive cell. Otherwise, the primitive cell standardized by
    spglib is used as the primitive cell.

    * `symmetrize_cell=False` and `find_primitive=False`.

    The unit cell and primitive cell are the same as the input cell.

    Parameters
    ----------
    cell_filename : Path
        Input cell file name.
    tolerance : float
        Symmetry search tolerance.
    symmetrize_cell : bool
        If True, a standardize unit cell is generated.
    find_primitive : bool
        If False, the standardized unit cell (``symmetrize_cell=True``) or the
        input cell (``symmetrize_cell=False``) is used as the primitive cell. If
        True, primitive cell is made from the standardized cell
        (``symmetrize_cell=True``) or the input cell
        (``symmetrize_cell=False``).
    primitive_cell_choice : PrimitiveCellChoice
        When `primitive_cell_choice==PrimitiveCellChoice.REDUCED` The primitive
        cell is reduced by spglib.

    """
    sym_dataset = get_symmetry_dataset(input_cell, tolerance=tolerance)
    if symmetrize_cell:
        unitcell, _primitive, tmat = generate_standardized_cells(
            sym_dataset, tolerance=tolerance
        )
        if find_primitive:
            primitive = _primitive
            if len(_primitive) != len(unitcell):
                click.echo(
                    "Transformation matrix from conventional unit cell to "
                    "prmitive cell is"
                )
                for v in tmat:
                    click.echo(f"  [{v[0]:6.3f} {v[1]:6.3f} {v[2]:6.3f}]")
        else:
            primitive = unitcell
    else:
        unitcell = input_cell
        _primitive, tmat = get_primitive_cell(
            unitcell, sym_dataset, tolerance=tolerance
        )
        if find_primitive:
            if len(_primitive) == len(input_cell):
                primitive = input_cell
                click.echo(
                    "Input cell was already a primitive cell, and it is used as "
                    "the primitive cell."
                )
            else:
                primitive = _primitive
                click.echo("Found a primitive cell whose transformation matrix is")
                for v in tmat:
                    click.echo(f"  [{v[0]:6.3f} {v[1]:6.3f} {v[2]:6.3f}]")
        else:
            primitive = unitcell
            if len(_primitive) == len(unitcell):
                pass
            else:
                click.echo(
                    "Input cell is not a primitive cell from the symmetry point of "
                    "view. "
                )
                click.echo(
                    "But velph will consider the input cell as the primitive cell. "
                )
                click.echo(
                    "For reference, below is a potential primitive cell found in "
                    "the input cell."
                )
                click.echo("-" * 70)
                click.echo("\n".join(get_vasp_structure_lines(_primitive)).strip())
                click.echo("-" * 70)

    click.echo("Supercell is generated with respect to the cell below.")
    click.echo("-" * 80)
    click.echo(str(unitcell))
    click.echo("-" * 80)

    if primitive_cell_choice is PrimitiveCellChoice.REDUCED:
        primitive = get_reduced_cell(primitive, tolerance=tolerance)

    return unitcell, primitive, sym_dataset


def _parse_velph_template(
    velph_template_fp: Optional[Union[str, bytes, os.PathLike, io.IOBase]],
) -> Optional[dict]:
    """Read velph-toml template file.

    The types str, bytes, and os.PathLike are used to represent file names,
    while toml_str does not serve this purpose. If there is a need to pass
    toml_str, it can be achieved by using io.BytesIO(toml_str.encode('utf-8')).

    """
    if velph_template_fp is None:
        return None

    if isinstance(velph_template_fp, io.IOBase):
        return tomli.load(velph_template_fp)
    else:
        with open(velph_template_fp, "rb") as f:
            template_dict = tomli.load(f)

        click.echo(f'Read velph template file "{velph_template_fp}".')
        return template_dict


def _get_velph_dict(
    template_dict: Optional[dict],
) -> dict:
    """Return velph_dict.

    velph_dict is made by default_template_dict, template_dict and
    schedular_template_dict.

    This function implements very human-like actions.

    * [vasp.{key}] is simply replaced by template_dict["vasp"][key] unless
      key=="incar".
    * [vasp.incar] is treated specially by merging but not replacing.

    """
    velph_dict = copy.deepcopy(default_template_dict)
    if template_dict is not None:
        _update_velph_dict_by_template_dict(velph_dict, template_dict)

    return velph_dict


def _update_velph_dict_by_template_dict(velph_dict: dict, template_dict: dict):
    """Update velph_dict by template dict.

    [phelel], [vasp], ...

    """
    for key in template_dict:
        if key in velph_dict:
            if key == "vasp":
                _update_vasp_dict_by_template_dict(velph_dict, template_dict)
            else:
                velph_dict[key] = template_dict[key]


def _update_vasp_dict_by_template_dict(velph_dict: dict, template_dict: dict):
    """Update vasp_dict by template dict.

    [vasp.incar], [vasp.calc_key.incar], [vasp.calc_key.kpoints], ...

    """
    vasp_incar = _merge_vasp_incar_section(velph_dict, template_dict)
    for calc_key in template_dict["vasp"]:
        if calc_key == "incar":
            continue
        if calc_key in velph_dict["vasp"]:
            _update_vasp_calc_types(
                velph_dict["vasp"][calc_key],
                template_dict["vasp"][calc_key],
            )
        else:
            velph_dict["vasp"][calc_key] = template_dict["vasp"][calc_key]
    velph_dict["vasp"]["incar"] = vasp_incar


def _merge_vasp_incar_section(velph_dict: dict, template_dict: dict) -> dict:
    """Merge [vasp.incar] sections in default and template dicts."""
    vasp_incar = velph_dict["vasp"]["incar"]
    if "incar" in template_dict["vasp"]:
        for key in template_dict["vasp"]["incar"]:
            vasp_incar[key] = template_dict["vasp"]["incar"][key]

    vasp_incar_str = "[vasp.incar] (basic INCAR settings)"
    special_tags = [f"  {key} = {val}" for key, val in vasp_incar.items()]
    click.echo("\n".join([f"{vasp_incar_str}"] + special_tags))

    return vasp_incar


def _update_vasp_calc_types(vasp_calc_dict: dict, template_calc_dict: dict):
    """Update [vasp.calc_type].

    Treatment of [vasp.calc_type.*].

    """
    for key, val in template_calc_dict.items():
        if key in vasp_calc_dict and isinstance(val, dict):
            if isinstance(vasp_calc_dict[key], dict):
                vasp_calc_dict[key].update(val)
                continue
            else:
                raise RuntimeError(
                    "velph template and default template are inconsistent."
                )
        vasp_calc_dict[key] = template_calc_dict[key]


def _get_toml_lines(
    velph_dict: dict,
    vip: VelphInitParams,
    unitcell: PhonopyAtoms,
    primitive: PhonopyAtoms,
    cell_choices: dict,
    sym_dataset: dict,
    phelel_dir_name: str = "phelel",
) -> list[str]:
    """Return velph-toml lines."""
    supercell_dimension = _get_supercell_dimension(
        velph_dict.get("phelel", {}),
        vip.max_num_atoms,
        sym_dataset,
        vip.find_primitive,
    )
    if supercell_dimension is None:
        click.echo("", err=True)
        click.echo("Error | Supercell size could not be determined.", err=True)
        click.echo(
            "      | Specify max_num_atoms or [phelel.supercell_dimension].",
            err=True,
        )
        return None

    phonon_supercell_dimension = _get_supercell_dimension(
        velph_dict.get("phonopy", {}),
        vip.phonon_max_num_atoms,
        sym_dataset,
        vip.find_primitive,
    )

    click.echo("[phelel]")
    _show_supercell_dimension(supercell_dimension)

    if phonon_supercell_dimension is not None:
        click.echo("[phonopy]")
        _show_supercell_dimension(phonon_supercell_dimension)

    (
        kpoints_dict,
        kpoints_dense_dict,
        qpoints_dict,
        kpoints_opt_dict,
    ) = _get_kpoints_dict(
        vip.kspacing,
        vip.kspacing_dense,
        vip.use_grg,
        vip.tolerance,
        unitcell,
        primitive,
        sym_dataset,
        supercell_dimension,
        phonon_supercell_dimension,
        cell_choices["nac"],
        cell_choices["relax"],
        phelel_dir_name=phelel_dir_name,
    )

    if "vasp" in velph_dict:
        _update_kpoints_by_vasp_dict(
            kpoints_dict,
            kpoints_dense_dict,
            qpoints_dict,
            kpoints_opt_dict,
            velph_dict["vasp"],
        )
        _show_kpoints_lines(
            kpoints_dict,
            kpoints_dense_dict,
            velph_dict["vasp"],
            vip.kspacing,
            vip.kspacing_dense,
        )

    #
    # velph.toml
    #
    lines = []

    # [phelel]
    if "phelel" in velph_dict:
        lines += _get_phelel_lines(
            velph_dict,
            supercell_dimension,
            primitive,
            vip.amplitude,
            vip.diagonal,
            vip.plusminus,
            vip.phelel_nosym,
        )

    # [phonopy]
    if phonon_supercell_dimension is not None:
        lines += ["[phonopy]"]
        lines += _get_supercell_dimension_lines(phonon_supercell_dimension)
        lines += _get_displacement_settings_lines(
            velph_dict, "phonopy", vip.amplitude, vip.diagonal, vip.plusminus
        )
        lines.append("")

    # [vasp.*]
    if "vasp" in velph_dict:
        lines += _get_vasp_lines(
            velph_dict["vasp"],
            kpoints_dict,
            kpoints_dense_dict,
            kpoints_opt_dict,
            qpoints_dict,
            cell_choices["nac"],
            cell_choices["relax"],
            phelel_dir_name=phelel_dir_name,
        )

    # [scheduler]
    if "scheduler" in velph_dict:
        lines.append("[scheduler]")
        lines.append(tomli_w.dumps(velph_dict["scheduler"]))

    # [symmetry]
    if sym_dataset is not None:
        lines.append("[symmetry]")
        spg_type = sym_dataset.international
        lines.append(f'spacegroup_type = "{spg_type}"')
        lines.append(f"tolerance = {vip.tolerance}")
        if len(unitcell) != len(primitive):
            pmat = (np.linalg.inv(unitcell.cell) @ primitive.cell).T
            lines.append("primitive_matrix = [")
            for v in pmat:
                lines.append(f"  [ {v[0]:18.15f}, {v[1]:18.15f}, {v[2]:18.15f} ],")
            lines.append("]")
            lines.append("")

    # [unitcell]
    lines += _get_cell_toml_lines(unitcell, cell_name="unitcell")

    # [primitive_cell]
    if primitive is not None:
        lines += _get_cell_toml_lines(primitive, cell_name="primitive_cell")

    return lines


def _get_fft_mesh(velph_dict: dict, primitive: PhonopyAtoms) -> Optional[np.ndarray]:
    """FFT mesh is computed from encut and prec in INCAR dict.

    Two possiblity of encut sourse, [vasp.selfenergy.incar] and [vasp.incar].

    """
    cutoff_eV = None
    prec = None

    incar_dict_selfenergy = _get_incar_dict_with_encut(
        velph_dict, ["vasp", "selfenergy", "incar"]
    )
    incar_dict = _get_incar_dict_with_encut(velph_dict, ["vasp", "incar"])
    incar_dict.update(incar_dict_selfenergy)
    if "encut" in incar_dict:
        cutoff_eV = incar_dict["encut"]
        if "prec" in incar_dict:
            prec = incar_dict["prec"]
        if cutoff_eV is not None:
            fft_mesh = CutoffToFFTMesh.get_FFTMesh(
                cutoff_eV, primitive.cell, incar_prec=prec
            )
            return fft_mesh
    return None


def _get_incar_dict_with_encut(velph_dict: dict, nested_keys: list[str]) -> dict:
    """Return INCAR dict containing encut and prec keys."""
    incar_dict = velph_dict
    for key in nested_keys:
        try:
            incar_dict = incar_dict[key]
        except KeyError:
            return {}

    return {
        key.lower(): val
        for key, val in incar_dict.items()
        if key.lower() in ("encut", "prec")
    }


def _get_kpoints_dict(
    vip_kspacing: float,
    vip_kspacing_dense: float,
    vip_use_grg: bool,
    vip_tolerance: float,
    unitcell: PhonopyAtoms,
    primitive: PhonopyAtoms,
    sym_dataset: SpglibDataset,
    supercell_dimension: np.ndarray,
    phonon_supercell_dimension: Optional[np.ndarray],
    cell_for_nac: CellChoice,
    cell_for_relax: CellChoice,
    phelel_dir_name: str = "phelel",
) -> tuple[dict, dict, dict, dict]:
    """Return kpoints dicts."""
    use_grg_unitcell = vip_use_grg and (len(unitcell) == len(primitive))
    sym_dataset_prim = get_symmetry_dataset(primitive, tolerance=vip_tolerance)

    # Grid matrix for unitcell
    gm = GridMatrix(
        2 * np.pi / vip_kspacing,
        lattice=unitcell.cell,
        symmetry_dataset=sym_dataset,
        use_grg=use_grg_unitcell,
    )

    # Grid matrix for primitive cell
    gm_prim = GridMatrix(
        2 * np.pi / vip_kspacing,
        lattice=primitive.cell,
        symmetry_dataset=sym_dataset_prim,
        use_grg=vip_use_grg,
    )

    # Grid matrix for supercell
    supercell = get_supercell(unitcell, supercell_dimension)
    sym_dataset_super = get_symmetry_dataset(supercell, tolerance=vip_tolerance)
    gm_super = GridMatrix(
        2 * np.pi / vip_kspacing,
        lattice=supercell.cell,
        symmetry_dataset=sym_dataset_super,
        use_grg=False,
    )

    # Grid matrix for phonon supercell
    if phonon_supercell_dimension is not None:
        phonon_supercell = get_supercell(unitcell, phonon_supercell_dimension)
        sym_dataset_super = get_symmetry_dataset(
            phonon_supercell, tolerance=vip_tolerance
        )
        gm_phonon_super = GridMatrix(
            2 * np.pi / vip_kspacing,
            lattice=phonon_supercell.cell,
            symmetry_dataset=sym_dataset_super,
            use_grg=False,
        )
    else:
        gm_phonon_super = gm_super

    # Dense grid matrix for primitive cell
    gm_dense_prim = GridMatrix(
        2 * np.pi / vip_kspacing_dense,
        lattice=primitive.cell,
        symmetry_dataset=sym_dataset_prim,
        use_grg=vip_use_grg,
    )

    # Build return values
    kpoints_dict = _get_kpoints_by_kspacing(
        gm,
        gm_prim,
        gm_super,
        gm_phonon_super,
        cell_for_nac,
        cell_for_relax,
        phelel_dir_name=phelel_dir_name,
    )
    kpoints_dense_dict = _get_kpoints_by_kspacing_dense(gm_dense_prim)
    qpoints_dict: dict = {}
    kpoints_opt_dict: dict = {}

    return kpoints_dict, kpoints_dense_dict, qpoints_dict, kpoints_opt_dict


def _update_kpoints_by_vasp_dict(
    kpoints_dict: dict,
    kpoints_dense_dict: dict,
    qpoints_dict: dict,
    kpoints_opt_dict: dict,
    vasp_dict: dict,
) -> None:
    """Overwrite kpoints_(dense_)dict if kpoints are defined in vasp_dict.

    "phelel.phonon" are "phono3py.phonon" should be understood as
    vasp_dict["phelel"]["phonon"] and vasp_dict["phono3py"]["phonon"].

    """
    for key in _get_calc_types(vasp_dict):
        calc_type_dict = _get_dict_by_split_dots(vasp_dict, key)
        if "kpoints" in calc_type_dict:
            kpoints_dict[key] = calc_type_dict["kpoints"]
        if "kpoints_dense" in calc_type_dict:
            kpoints_dense_dict[key] = calc_type_dict["kpoints_dense"]
        if "qpoints" in calc_type_dict:
            qpoints_dict[key] = calc_type_dict["qpoints"]
        if "kpoints_opt" in calc_type_dict:
            kpoints_opt_dict[key] = calc_type_dict["kpoints_opt"]


def _get_calc_types(vasp_dict: dict) -> list[str]:
    keys = list(vasp_dict.keys())
    for key in ["phelel.phonon", "phono3py.phonon"]:
        if _get_dict_by_split_dots(vasp_dict, key):
            keys.append(key)
    return keys


def _get_dict_by_split_dots(dict_in: dict, string: str) -> dict:
    """Return dict by splitting string."""
    keys = string.split(".")
    _dict_in = dict_in
    for key in keys:
        _dict_in = _dict_in.get(key, {})
    return _dict_in


def _show_kpoints_lines(
    kpoints_dict: dict,
    kpoints_dense_dict: dict,
    vasp_dict: dict,
    vip_kspacing: float,
    vip_kspacing_dense: float,
):
    k_mesh_lines = [f"[vasp.*.kpoints.mesh] (*kspacing={vip_kspacing})"]
    for key in _get_calc_types(vasp_dict):
        if key == "ph_bands":
            continue
        if key in kpoints_dict:
            mesh = kpoints_dict[key].get("mesh")
            if mesh is not None:
                line = f"  {key}: {np.array(mesh)}"
                if "kpoints" not in _get_dict_by_split_dots(vasp_dict, key):
                    line += "*"
                k_mesh_lines.append(line)
    k_mesh_lines.append(
        f"[vasp.*.kpoints_dense.mesh] (*kspacing_dense={vip_kspacing_dense})"
    )
    for key in vasp_dict:
        if key in kpoints_dense_dict:
            mesh = kpoints_dense_dict[key].get("mesh")
            if mesh is not None:
                line = f"  {key}: {np.array(mesh)}"
                if "kpoints_dense" not in vasp_dict[key]:
                    line += "*"
                k_mesh_lines.append(line)
    if k_mesh_lines:
        click.echo("\n".join(k_mesh_lines))


def _get_kpoints_by_kspacing(
    gm: GridMatrix,
    gm_prim: GridMatrix,
    gm_super: GridMatrix,
    gm_phonon_super: Optional[GridMatrix],
    cell_for_nac: CellChoice,
    cell_for_relax: CellChoice,
    phelel_dir_name: str = "phelel",
) -> dict:
    """Return kpoints dict.

    Parameters
    ----------
    gm : GridMatrix
        Grid matrix of unit cell.
    gm_prim : GridMatrix
        Grid matrix of primitive cell. The primitive cell can be a reduced cell.
    gm_super : GridMatrix
        Grid matrix of supercell.
    cell_for_nac : CellChoice
        Cell choice for NAC calculation among unit cell or primitive cell. The
        primitive cell can be a reduced cell.
    cell_for_relax : CellChoice
        Cell choice for relax calculation among unit cell or primitive cell. The
        primitive cell can be a reduced cell.

    Returns
    -------
    dict
        kpoints information for each calculation type is stored. The keys must
        be calc_type names such as "phelel", "phelel.phonon", "selfenergy", etc.

    """
    if gm_super.grid_matrix is None:
        supercell_kpoints = {"mesh": gm_super.D_diag}
    else:
        supercell_kpoints = {"mesh": gm_super.grid_matrix}
    if gm_phonon_super is not None:
        if gm_super.grid_matrix is None:
            phonon_supercell_kpoints = {"mesh": gm_phonon_super.D_diag}
        else:
            phonon_supercell_kpoints = {"mesh": gm_phonon_super.grid_matrix}
    else:
        phonon_supercell_kpoints = None
    if gm_prim.grid_matrix is None:
        selfenergy_kpoints = {"mesh": gm_prim.D_diag}
        el_bands_kpoints = {"mesh": gm_prim.D_diag}
    else:
        selfenergy_kpoints = {"mesh": gm_prim.grid_matrix}
        el_bands_kpoints = {"mesh": gm_prim.grid_matrix}
    ph_bands_kpoints = {"mesh": [1, 1, 1]}

    # nac
    if cell_for_nac is CellChoice.UNITCELL:
        gm_nac = gm
    elif cell_for_nac is CellChoice.PRIMITIVE:
        gm_nac = gm_prim
    else:
        raise RuntimeError("This is something that sholud not happen.")
    if gm_nac.grid_matrix is None:
        nac_kpoints = {"mesh": gm_nac.D_diag}
    else:
        nac_kpoints = {"mesh": gm_nac.grid_matrix}

    # relax
    if cell_for_relax is CellChoice.UNITCELL:
        gm_relax = gm
    elif cell_for_relax is CellChoice.PRIMITIVE:
        gm_relax = gm_prim
    else:
        raise RuntimeError("This is something that sholud not happen.")
    if gm_relax.grid_matrix is None:
        relax_kpoints = {"mesh": gm_relax.D_diag}
    else:
        relax_kpoints = {"mesh": gm_relax.grid_matrix}

    # keys are calc_types.
    return {
        phelel_dir_name: supercell_kpoints,
        "phono3py": supercell_kpoints,
        "phelel.phonon": phonon_supercell_kpoints,
        "phono3py.phonon": phonon_supercell_kpoints,
        "selfenergy": selfenergy_kpoints,
        "transport": selfenergy_kpoints,
        "relax": relax_kpoints,
        "nac": nac_kpoints,
        "el_bands": el_bands_kpoints,
        "ph_bands": ph_bands_kpoints,
    }


def _get_kpoints_by_kspacing_dense(gm_dense_prim: GridMatrix) -> dict:
    """Return kpoints dict of dense grid.

    Returns
    -------
    dict
        kpoints information for each calculation type is stored. The keys must
        be calc_type names such as "selfenergy", etc.

    """
    if gm_dense_prim.grid_matrix is None:
        selfenergy_kpoints_dense = {"mesh": gm_dense_prim.D_diag}
        el_bands_kpoints_dense = {"mesh": gm_dense_prim.D_diag}
    else:
        selfenergy_kpoints_dense = {"mesh": gm_dense_prim.grid_matrix}
        el_bands_kpoints_dense = {"mesh": gm_dense_prim.grid_matrix}
    # keys are calc_types.
    return {
        "selfenergy": selfenergy_kpoints_dense,
        "transport": selfenergy_kpoints_dense,
        "el_bands": el_bands_kpoints_dense,
    }


def _get_vasp_lines(
    vasp_dict: dict,
    kpoints_dict: dict,
    kpoints_dense_dict: dict,
    kpoints_opt_dict: dict,
    qpoints_dict: dict,
    cell_for_nac: CellChoice,
    cell_for_relax: CellChoice,
    phelel_dir_name: str = "phelel",
) -> list:
    incar_commons = _get_incar_commons(vasp_dict)

    lines = []

    for calc_type in (phelel_dir_name, "phono3py"):
        if calc_type in vasp_dict:
            _vasp_dict = vasp_dict[calc_type]
            _add_incar_lines(lines, _vasp_dict, incar_commons, calc_type)
            lines.append(f"[vasp.{calc_type}.kpoints]")
            _add_kpoints_lines(lines, kpoints_dict[calc_type])
            _add_calc_type_scheduler_lines(lines, _vasp_dict, calc_type)
            lines.append("")

    for calc_type in (phelel_dir_name, "phono3py"):
        if calc_type in vasp_dict:
            if "phonon" in vasp_dict[calc_type]:
                _vasp_dict = vasp_dict[calc_type]["phonon"]
                _add_incar_lines(
                    lines, _vasp_dict, incar_commons, f"{calc_type}.phonon"
                )
                lines.append(f"[vasp.{calc_type}.phonon.kpoints]")
                _add_kpoints_lines(lines, kpoints_dict[f"{calc_type}.phonon"])
                _add_calc_type_scheduler_lines(lines, _vasp_dict, f"{calc_type}.phonon")
                lines.append("")

    for calc_type in ("selfenergy", "transport"):
        if calc_type in vasp_dict:
            # primitive cell
            _vasp_dict = vasp_dict[calc_type]
            _add_incar_lines(lines, _vasp_dict, incar_commons, calc_type)
            lines.append(f"[vasp.{calc_type}.kpoints]")
            _add_kpoints_lines(lines, kpoints_dict[calc_type])
            lines.append(f"[vasp.{calc_type}.kpoints_dense]")
            _add_kpoints_lines(lines, kpoints_dense_dict[calc_type])
            _add_calc_type_scheduler_lines(lines, _vasp_dict, calc_type)
            lines.append("")

    if "relax" in vasp_dict:
        lines.append("[vasp.relax]")
        assert cell_for_relax in CellChoice
        assert cell_for_relax is not CellChoice.UNSPECIFIED
        _vasp_dict = vasp_dict["relax"]
        lines.append(f'cell = "{cell_for_relax.value}"')
        _add_incar_lines(lines, _vasp_dict, incar_commons, "relax")
        lines.append("[vasp.relax.kpoints]")
        _add_kpoints_lines(lines, kpoints_dict["relax"])
        _add_calc_type_scheduler_lines(lines, _vasp_dict, "relax")
        lines.append("")

    if "nac" in vasp_dict:
        lines.append("[vasp.nac]")
        assert cell_for_nac in CellChoice
        assert cell_for_nac is not CellChoice.UNSPECIFIED
        _vasp_dict = vasp_dict["nac"]
        lines.append(f'cell = "{cell_for_nac.value}"')
        _add_incar_lines(lines, _vasp_dict, incar_commons, "nac")
        lines.append("[vasp.nac.kpoints]")
        _add_kpoints_lines(lines, kpoints_dict["nac"])
        _add_calc_type_scheduler_lines(lines, _vasp_dict, "nac")
        lines.append("")

    if "el_bands" in vasp_dict:
        # primitive cell
        el_bands_kpoints_dense = kpoints_dense_dict.get("el_bands")
        el_bands_kpoints_opt = kpoints_opt_dict.get("el_bands")
        _vasp_dict = vasp_dict["el_bands"]
        _add_incar_lines(lines, _vasp_dict, incar_commons, "el_bands")
        lines.append("[vasp.el_bands.kpoints]")
        _add_kpoints_lines(lines, kpoints_dict["el_bands"])
        if el_bands_kpoints_opt:
            lines.append("[vasp.el_bands.kpoints_opt]")
            _add_kpoints_lines_bands(lines, el_bands_kpoints_opt)
        if el_bands_kpoints_dense:
            lines.append("[vasp.el_bands.kpoints_dense]")
            _add_kpoints_lines(lines, el_bands_kpoints_dense)
        _add_calc_type_scheduler_lines(lines, _vasp_dict, "el_bands")
        lines.append("")

    if "ph_bands" in vasp_dict:
        # primitive cell
        ph_bands_qpoints = qpoints_dict.get("ph_bands")
        _vasp_dict = vasp_dict["ph_bands"]
        _add_incar_lines(lines, _vasp_dict, incar_commons, "ph_bands")
        lines.append("[vasp.ph_bands.kpoints]")
        _add_kpoints_lines(lines, kpoints_dict["ph_bands"])
        if ph_bands_qpoints:
            lines.append("[vasp.ph_bands.qpoints]")
            _add_kpoints_lines_bands(lines, ph_bands_qpoints)
        _add_calc_type_scheduler_lines(lines, _vasp_dict, "ph_bands")
        lines.append("")

    return lines


def _get_incar_commons(vasp_dict: dict) -> dict:
    """Extract common INCAR parameters.

    Parameters written in ``vasp_dict['incar']`` are considered as common
    INCAR parameters.

    Default common INCAR parameters are hard coded in ``default_template_dict``.

    """
    assert "incar" in vasp_dict
    return vasp_dict["incar"]


def _merge_incar_commons(incar: dict, incar_commons: dict):
    """Merge INCAR parameters in template and common INCAR parameters."""
    incar_copy = copy.deepcopy(incar)
    for key in incar_commons:
        if key not in incar_copy:
            incar_copy[key] = incar_commons[key]
    return {key: value for key, value in incar_copy.items() if value is not None}


def _get_phelel_lines(
    velph_dict: dict,
    supercell_dimension: np.ndarray,
    primitive: PhonopyAtoms,
    amplitude: Optional[float],
    diagonal: Optional[bool],
    plusminus: Optional[bool],
    phelel_nosym: bool,
) -> list:
    lines = []
    lines.append("[phelel]")
    lines.append(f'version = "{__version__}"')

    lines += _get_supercell_dimension_lines(supercell_dimension)
    lines += _get_displacement_settings_lines(
        velph_dict, "phelel", amplitude, diagonal, plusminus
    )

    if phelel_nosym:
        lines.append("nosym = true")

    fft_mesh = _get_fft_mesh(velph_dict, primitive)
    try:
        if fft_mesh is None and "fft_mesh" in velph_dict["phelel"]:
            fft_mesh = velph_dict["phelel"]["fft_mesh"]
        if fft_mesh is not None:
            lines.append("fft_mesh = [{:d}, {:d}, {:d}]".format(*fft_mesh))
    except KeyError:
        pass
    lines.append("")
    return lines


def _add_incar_lines(lines: list, vasp_dict: dict, incar_commons: dict, calc_type: str):
    if "incar" in vasp_dict:
        lines.append(f"[vasp.{calc_type}.incar]")
        lines.append(
            tomli_w.dumps(
                _merge_incar_commons(vasp_dict["incar"], incar_commons)
            ).strip()
        )


def _add_kpoints_lines_bands(lines: list, kpt_dict: dict) -> None:
    lines.append("line = {:d}".format(kpt_dict["line"]))
    if "path" in kpt_dict:
        lines.append("path = " + str(kpt_dict["path"]))
        points = kpt_dict["path"]
        for i in np.unique(points):
            lines.append(
                str(i) + " = " + "[{:f}, {:f}, {:f}]".format(*kpt_dict[str(i)])
            )


def _add_kpoints_lines(lines: list, kpt_dict: dict) -> None:
    if len(np.ravel(kpt_dict["mesh"])) == 3:
        lines.append("mesh = [{:d}, {:d}, {:d}]".format(*kpt_dict["mesh"]))
    elif len(np.ravel(kpt_dict["mesh"])) == 9:
        lines.append("mesh = [")
        lines.append("  [{:d}, {:d}, {:d}],".format(*kpt_dict["mesh"][0]))
        lines.append("  [{:d}, {:d}, {:d}],".format(*kpt_dict["mesh"][1]))
        lines.append("  [{:d}, {:d}, {:d}]".format(*kpt_dict["mesh"][2]))
        lines.append("]")
    if "shift" in kpt_dict:
        lines.append("shift = [{:f}, {:f}, {:f}]".format(*kpt_dict["shift"]))


def _add_calc_type_scheduler_lines(lines: list, vasp_dict: dict, calc_type: str):
    if "scheduler" in vasp_dict:
        lines.append(f"[vasp.{calc_type}.scheduler]")
        lines.append(tomli_w.dumps(vasp_dict["scheduler"]).strip())


def _get_cell_toml_lines(
    unitcell: PhonopyAtoms, cell_name: str = "cell", show_masses: bool = False
) -> list:
    """Return crystal structure lines in toml.

    Masses are not presented by default because they are inconsistent with those
    in VASP.

    """
    if cell_name:
        lines = [f"[{cell_name}]"]
    lines.append("lattice = [")
    for v, a in zip(unitcell.cell, ("a", "b", "c")):
        lines.append("  [ %21.15f, %21.15f, %21.15f ], # %s" % (v[0], v[1], v[2], a))
    lines.append("]")
    if unitcell.masses is None:
        masses = [None] * len(unitcell.symbols)
    else:
        masses = unitcell.masses
    if unitcell.magnetic_moments is None:
        magnetic_moments = [None] * len(unitcell.symbols)
    else:
        magnetic_moments = unitcell.magnetic_moments
    for i, (s, v, m, mag) in enumerate(
        zip(unitcell.symbols, unitcell.scaled_positions, masses, magnetic_moments)
    ):
        lines.append(f"[[{cell_name}.points]]  # {i + 1}")
        lines.append(f'symbol = "{s}"')
        lines.append(f"coordinates = [ {v[0]:18.15f}, {v[1]:18.15f}, {v[2]:18.15f} ]")
        if show_masses and m is not None:
            lines.append(f"mass = {m:f}")
        if mag is not None:
            if mag.ndim == 0:
                mag_str = f"{mag:.8f}"
            else:
                mag_str = f"[ {mag[0]:.8f}, {mag[1]:.8f}, {mag[2]:.8f} ]"
            lines.append(f"magnetic_moment = {mag_str}")
    return lines


def _show_supercell_dimension(dim: np.ndarray) -> None:
    key = "supercell_dimension"
    if np.array_equal(dim, np.diag(dim.diagonal())):
        click.echo(f"  {key}: {dim.diagonal()}")
    else:
        click.echo(f"  {key}:")
        for v in dim:
            click.echo(f"    {v}")


def _get_supercell_dimension_lines(
    supercell_dimension: Union[np.ndarray, Sequence],
) -> list:
    lines = []
    if (np.diag(np.diag(supercell_dimension)) == supercell_dimension).all():
        lines.append(
            "supercell_dimension = [{:d}, {:d}, {:d}]".format(
                *np.diag(supercell_dimension)
            )
        )
    else:
        fmt_str = (
            "supercell_dimension = "
            "[[{:d}, {:d}, {:d}], [{:d}, {:d}, {:d}], [{:d}, {:d}, {:d}]]"
        )
        lines.append(fmt_str.format(*np.ravel(supercell_dimension)))
    return lines


def _get_displacement_settings_lines(
    velph_dict: dict,
    calc_type: Literal["phelel", "phonopy"],
    amplitude: Optional[float],
    diagonal: Optional[bool],
    plusminus: Optional[bool],
) -> list:
    lines = []
    toml_dict = velph_dict.get(calc_type, {})
    if "amplitude" in toml_dict:
        lines.append(f'amplitude = {toml_dict["amplitude"]}')
    else:
        lines.append(f"amplitude = {amplitude}")

    if "diagonal" in toml_dict:
        _diagonal = toml_dict["diagonal"]
    else:
        _diagonal = diagonal
    assert isinstance(_diagonal, bool)
    if _diagonal:
        lines.append("diagonal = true")
    else:
        lines.append("diagonal = false")

    if "plusminus" in toml_dict:
        _plusminus = toml_dict["plusminus"]
    else:
        if plusminus is False:
            _plusminus = "auto"
        else:
            _plusminus = True
    if isinstance(_plusminus, bool):
        if _plusminus:
            lines.append("plusminus = true")
        else:
            lines.append("plusminus = false")
    elif isinstance(_plusminus, str):
        if _plusminus == "auto":
            lines.append('plusminus = "auto"')
        else:  # Fall back to default
            lines.append("plusminus = true")
    return lines



"""Templates of vasp inputs written in velph.toml.

``template_dict`` is used to generate ``velph.toml``, and these have a similar
dictionary structure.

"phelel":
    General parameters to run phelel. "supercell_dimension":
        list[list[int, int, int]], list[int, int, int], or int Supercell
        dimension with respect to unit cell. Typically given by three integers,
        but nine integers can be used. If one integer value is given, supercell
        dimension is estimated by taking this value as the maximum number of
        atoms in the supercell.
    "fft_mesh":
        FFT mesh used in primitive cell for sandwich.
"vasp":
    Types of VASP calculations, e.g., "phelel", "relax", "nac", "selfenergy",
    "transport", and default set of INCAR settins, "incar".

    The default incar settings in the "incar" section are copied to the
    "incar" sections of all VASP calculation types if those default INCAR
    settings are not available in each section. To prevent from writing
    default INCAR settings, i.e., avoiding unnecessary INCAR settings in a
    specific VASP calculation type, INCAR tag with the value of ``None`` has to
    be explicitly described (see "ph_bands" section.)

    In each VASP calculation following settings are specified.
        "incar":
            VASP INCAR setting by a dictionary.
        "kpoints":
            Sampling mesh of k-points. "mesh": list[list[int, int, int]],
            list[int, int, int]
                Typically three integers. With nine integers, generalized
                regular grid is used.
            "shift": list[float, float, float]
                Shift of mesh with respect to neighboring grid points. 0.5 means
                half-grid shift.
            "kspacing": float
                This value is treated as same as KSPACING with KGAMMA=.TRUE. in
                VASP INCAR setting.
        "kpoints_dense":
            This is used for sandwich. Charge density calculation is performed
            using the configuration of "kpoints".
"scheduler":
    Schedular configuration. Values of the kyes are inserted to specific
    schedular templeate found in ``velph/template/scheduler``. This section is
    replaced by ``.config/velph/scheduler.toml`` if this path exists.

"""

from __future__ import annotations

from typing import Any

default_template_dict: dict[str, Any] = {
    "phelel": {},
    "phonopy": {},
    "phono3py": {},
    "vasp": {
        "incar": {
            "ismear": 0,
            "sigma": 0.01,
            "ediff": 1e-8,
            "encut": 500,
            "prec": "accurate",
            "lreal": False,
            "lwave": False,
            "lcharg": False,
        },
        "selfenergy": {
            "incar": {
                "elph_run": True,
                "elph_selfen_fan": True,
                "elph_selfen_dw": True,
                "elph_selfen_delta": 0.01,
                "elph_selfen_temps": [300],
                "elph_wf_cache_mb": 1000,
                "elph_wf_cache_prefill": True,
                "elph_wf_redistribute": True,
                "elph_nbands": 100,
                "elph_nbands_sum": [50, 100],
                "elph_selfen_gaps": True,
                "elph_ismear": -24,
                "elph_fermi_nedos": 21,
            },
        },
        "transport": {
            "incar": {
                "elph_fermi_nedos": 5001,
                "elph_ismear": -24,
                "elph_mode": "transport",
                "elph_selfen_carrier_den": 0.0,
                "elph_scattering_approx": ["serta", "mrta_lambda"],
                "elph_selfen_temps": [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    9,
                    10,
                    20,
                    30,
                    40,
                    50,
                    60,
                    70,
                    80,
                    90,
                    100,
                    110,
                    120,
                    130,
                    140,
                    150,
                    160,
                    170,
                    180,
                    190,
                    200,
                    210,
                    220,
                    230,
                    240,
                    250,
                    260,
                    270,
                    280,
                    290,
                    300,
                    400,
                    500,
                    600,
                    700,
                    800,
                    900,
                    1000,
                    1100,
                    1200,
                    1300,
                    1400,
                ],
                "elph_transport_nedos": 501,
                "elph_wf_redistribute": True,
                "elph_wf_redistribute_opt": 0,
                "elph_wf_comm_opt": 0,
            },
        },
        "el_bands": {
            "incar": {"ibrion": -1, "nsw": 0, "lorbit": 11, "nedos": 5001},
            "kpoints_opt": {
                "line": 51,
            },
        },
        "ph_bands": {
            "incar": {
                "ibrion": -1,
                "nsw": 0,
                "elph_run": True,
                "ismear": None,
                "sigma": None,
                "ediff": None,
                "lreal": None,
                "lwave": None,
                "lcharg": None,
            },
            "qpoints": {
                "line": 51,
            },
        },
        "phonopy": {
            "incar": {
                "addgrid": True,
                "isym": 0,
            },
        },
        "phono3py": {
            "incar": {
                "addgrid": True,
                "isym": 0,
            },
        },
        "phelel": {
            "incar": {
                "lwap": True,
                "isym": 0,
            },
        },
        "supercell": {  # This is deprecated. Use "phelel" instead.
            "incar": {
                "lwap": True,
                "isym": 0,
            },
        },
        "relax": {
            "incar": {
                "ediffg": -1e-6,
                "ibrion": 2,
                "isif": 3,
                "nsw": 10,
            },
        },
        "nac": {
            "incar": {
                "lepsilon": True,
                "npar": None,
                "ncore": None,
                "kpar": None,
            },
        },
    },
    "scheduler": {
        "scheduler_name": "(sge, etc, custom)",
        "job_name": "vasp-elph",
        "mpirun_command": "mpirun",
        "vasp_binary": "vasp_std",
        "prepend_text": "",
        "append_text": "",
        "custom_template": "",
    },
}



"""Implementation of velph-nac-generate."""

import pathlib
import shutil

import click
import tomli
from phonopy.interface.calculator import write_crystal_structure

from phelel.velph.cli.utils import (
    assert_kpoints_mesh_symmetry,
    choose_cell_in_dict,
    get_scheduler_dict,
    write_incar,
    write_kpoints_mesh_mode,
    write_launch_script,
)


def write_input_files(toml_filename: pathlib.Path) -> None:
    """Generate VASP nac inputs."""
    with open(toml_filename, "rb") as f:
        toml_dict = tomli.load(f)

    directory_name = "nac"
    directory = pathlib.Path(directory_name)
    directory.mkdir(parents=True, exist_ok=True)

    # POSCAR
    cell = choose_cell_in_dict(toml_dict, toml_filename, "nac")
    write_crystal_structure(directory / "POSCAR", cell)

    # INCAR
    write_incar(toml_dict["vasp"]["nac"]["incar"], directory, cell=cell)

    # KPOINTS
    kpoints_dict = toml_dict["vasp"]["nac"]["kpoints"]
    assert_kpoints_mesh_symmetry(toml_dict, kpoints_dict, cell)
    write_kpoints_mesh_mode(
        toml_dict["vasp"]["nac"]["incar"],
        directory,
        "vasp.nac.kpoints",
        toml_dict["vasp"]["nac"]["kpoints"],
    )

    # POTCAR
    if pathlib.Path("POTCAR").exists():
        shutil.copy2(pathlib.Path("POTCAR"), directory / "POTCAR")

    # Scheduler launch script
    if "scheduler" in toml_dict:
        scheduler_dict = get_scheduler_dict(toml_dict, "nac")
        write_launch_script(scheduler_dict, directory)

    click.echo(f'VASP input files were made in "{directory_name}".')



"""Calculation of Eliashberg function."""

from collections.abc import Sequence
from typing import Optional, Union

import h5py
import numpy as np
from phono3py.file_IO import get_filename_suffix
from phonopy import Phonopy
from phonopy.phonon.tetrahedron_mesh import TetrahedronMesh
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.grid_points import GridPoints
from phonopy.units import THzToEv


class EliashbergFunction:
    """Eliashberg function class."""

    def __init__(
        self,
        mesh: Union[Sequence, np.ndarray],
        primitive: PhonopyAtoms,
        rotations: np.ndarray,
        gamma: np.ndarray,
        frequencies: np.ndarray,
        eigenvalues: np.ndarray,
        fermi_level: Optional[float] = None,
    ):
        """Init method."""
        self._mesh = mesh
        self._primitive = primitive
        self._frequencies = frequencies
        self._eigenvalues = eigenvalues
        self._gamma = gamma
        self._fermi_level = fermi_level

        self._gp = GridPoints(
            self._mesh,
            np.linalg.inv(self._primitive.cell),
            is_gamma_center=True,
            rotations=rotations,
        )
        assert len(self._gp.ir_grid_points) == len(frequencies)
        self._eliashberg = None

        # 2pi cancels to convert unit of delta function
        self._unit_conversion = 1.0 / THzToEv / 2 / np.prod(self._mesh)

        if self._fermi_level is not None:
            self.run_electronic_DOS(np.array([self._fermi_level], dtype="double"))
            self._unit_conversion /= self._eldos[0]

    @property
    def eliashberg(self):
        """Return Eliashberg function."""
        return self._eliashberg

    @property
    def gp(self):
        """Return GridPoints instance."""
        return self._gp

    @property
    def electronic_dos(self):
        """Return electronic DOS."""
        return self._eldos

    def run_tetrahedron_method(self, frequency_points: Union[Sequence, np.ndarray]):
        """Calculate Eliashberg function spectrum using tetrahedron method."""
        thm = TetrahedronMesh(
            self._primitive,
            self._frequencies,
            self._mesh,
            self._gp.grid_address,
            self._gp.grid_mapping_table,
            self._gp.ir_grid_points,
        )
        self._eliashberg = np.zeros(len(frequency_points), dtype="double")
        thm.set(value="I", frequency_points=frequency_points)
        coef = self._gamma / self._frequencies
        for i, iw in enumerate(thm):
            # iw (fpoints, bands)
            if i == 0:
                continue
            w = self._gp.weights[i]
            self._eliashberg += np.dot(iw, coef[i]) * w
        self._eliashberg *= self._unit_conversion

    def run_electronic_DOS(self, frequency_points: Union[Sequence, np.ndarray]):
        """Calculate electronic DOS using tetrahedron method."""
        thm = TetrahedronMesh(
            self._primitive,
            self._eigenvalues,
            self._mesh,
            self._gp.grid_address,
            self._gp.grid_mapping_table,
            self._gp.ir_grid_points,
        )
        self._eldos = np.zeros(len(frequency_points), dtype="double")
        thm.set(value="I", frequency_points=frequency_points)
        for i, iw in enumerate(thm):
            # iw (fpoints, bands)
            if i == 0:
                continue
            w = self._gp.weights[i]
            self._eldos += iw.sum(axis=1) * w


def _collect_frequency_and_gamma(phonon: Phonopy):
    frequencies = phonon.get_mesh_dict()["frequencies"]
    ir_grid_points = phonon.mesh.ir_grid_points
    assert len(ir_grid_points) == len(frequencies)
    gamma = np.zeros_like(frequencies)
    for i, gp in enumerate(ir_grid_points):
        suffix = get_filename_suffix(phonon.mesh_numbers, grid_point=gp)
        filename = "gamma%s.hdf5" % suffix
        with h5py.File(filename, "r") as f:
            gamma[i] = f["gamma"][:]

    return frequencies, gamma


def _get_eigenvalues(mesh_numbers: np.ndarray, ir_grid_points: np.ndarray):
    suffix = get_filename_suffix(mesh_numbers)
    filename = "electron%s.hdf5" % suffix
    with h5py.File(filename, "r") as f:
        eigenvalues = f["eigenvalues"][:]
    return np.array(eigenvalues[ir_grid_points, :, 0], order="C", dtype="double")


def show_Eliashberg_function(
    phonon: Phonopy,
    num_frequency_points: int = 4001,
    fermi_level: Optional[float] = None,
):
    """Show Eliashberg functions at frequency points."""
    frequencies, gamma = _collect_frequency_and_gamma(phonon)
    ir_grid_points = phonon.mesh.ir_grid_points
    eigenvalues = _get_eigenvalues(phonon.mesh_numbers, ir_grid_points)
    ef = EliashbergFunction(
        phonon.mesh_numbers,
        phonon.primitive,
        phonon.primitive_symmetry.get_pointgroup_operations(),
        gamma,
        frequencies,
        eigenvalues,
        fermi_level=fermi_level,
    )
    print("# DOS at fermi level: %f" % ef.electronic_dos[0])
    fpts = np.linspace(
        frequencies.min() - 0.1, frequencies.max() + 0.1, num_frequency_points
    )
    ef.run_tetrahedron_method(fpts)
    for fp, dos in zip(fpts, ef.eliashberg):
        print(fp, dos)

    # To plot electronic DOS
    # fpts = np.linspace(eigenvalues.min() - 0.1, eigenvalues.max() + 0.1,
    #                    num_frequency_points)
    # ef.run_electronic_DOS(fpts)
    # print("")
    # print("")
    # for fp, dos in zip(fpts, ef.electronic_dos * 2):
    #     print(fp, dos)


def show_frequency_and_gamma(phonon: Phonopy):
    """Show frequecnies and Eliashberg functions."""
    frequencies, gamma = _collect_frequency_and_gamma(phonon)
    for f_band, g_band in zip(frequencies, gamma):
        for f, g in zip(f_band, g_band):
            print(f, g)



"""Calculation of lattice points."""

import numpy as np
from phonopy.structure.cells import SNF3x3


def get_lattice_points(p2s_matrix: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """Return lattice points and D (diagonal elements) of SNF."""
    snf = SNF3x3(p2s_matrix)
    snf.run()
    P_inv = np.rint(np.linalg.inv(snf.P)).astype(int)
    multi = np.diagonal(snf.D)
    lattice_points = np.array(
        np.dot(list(np.ndindex(tuple(multi))), P_inv.T), dtype="int_", order="C"
    )
    return lattice_points, multi



"""Implementation of velph-supercell-init."""

from __future__ import annotations

import pathlib
from typing import Optional

import click
import numpy as np
from phonopy.structure.atoms import parse_cell_dict

from phelel import Phelel
from phelel.cui.create_supercells import generate_phelel_supercells
from phelel.velph.cli.utils import get_nac_params


def run_init(
    toml_dict: dict,
    current_directory: pathlib.Path = pathlib.Path(""),
) -> Optional[Phelel]:
    """Generate supercell and displacements.

    current_directory : Path
        Used for test.

    """
    convcell = parse_cell_dict(toml_dict["unitcell"])
    supercell_matrix = toml_dict["phelel"].get("supercell_dimension", None)
    phonon_supercell_matrix = toml_dict["phelel"].get(
        "phonon_supercell_dimension", None
    )
    if "primitive_cell" in toml_dict:
        primitive = parse_cell_dict(toml_dict["primitive_cell"])
        primitive_matrix = np.dot(np.linalg.inv(convcell.cell.T), primitive.cell.T)
    else:
        primitive = convcell
        primitive_matrix = None

    is_symmetry = True
    try:
        if toml_dict["phelel"]["nosym"] is True:
            is_symmetry = False
    except KeyError:
        pass

    phe = Phelel(
        convcell,
        supercell_matrix=supercell_matrix,
        phonon_supercell_matrix=phonon_supercell_matrix,
        primitive_matrix=primitive_matrix,
        is_symmetry=is_symmetry,
        calculator="vasp",
    )

    is_diagonal = toml_dict["phelel"].get("diagonal", True)
    is_plusminus = toml_dict["phelel"].get("plusminus", "auto")
    amplitude = toml_dict["phelel"].get("amplitude", None)

    generate_phelel_supercells(
        phe,
        interface_mode="vasp",
        distance=amplitude,
        is_plusminus=is_plusminus,
        is_diagonal=is_diagonal,
    )

    nac_directory = current_directory / "nac"
    if nac_directory.exists():
        click.echo('Found "nac" directory. Read NAC params.')
        vasprun_path = nac_directory / "vasprun.xml"
        if vasprun_path.exists():
            nac_params = get_nac_params(
                toml_dict,
                vasprun_path,
                primitive,
                convcell,
                is_symmetry,
            )
            if nac_params is not None:
                phe.nac_params = nac_params
        else:
            click.echo('Not found "nac/vasprun.xml". NAC params were not included.')

    return phe



"""Implementation of velph-el_bands-plot."""

from __future__ import annotations

import pathlib

import click
import h5py
import matplotlib.pyplot as plt

from phelel.velph.cli.utils import (
    get_distances_along_BZ_path,
    get_reclat_from_vaspout,
    get_special_points,
)


def plot_el_bandstructures(
    window: tuple[float, float],
    vaspout_filename_bands: pathlib.Path,
    vaspout_filename_dos: pathlib.Path,
    plot_filename: pathlib.Path = pathlib.Path("el_bands/el_bands.pdf"),
):
    """Plot electronic band structure.

    Parameters
    ----------
    window : tuple(float, float)
        Energy window to draw band structure in eV with respect to Fermi level.
    vaspout_filename_bands : pathlib.Path
        Filename of vaspout.h5 of band structure.
    vaspout_filename_dos : pathlib.Path
        Filename of vaspout.h5 of DOS.
    plot_filename : pathlib.Path, optional
        File name of band structure plot.

    """
    f_h5py_bands = h5py.File(vaspout_filename_bands)
    f_h5py_dos = h5py.File(vaspout_filename_dos)
    efermi = f_h5py_dos["results"]["electron_dos_kpoints_opt"]["efermi"][()]
    emin = window[0]
    emax = window[1]
    _, axs = plt.subplots(1, 2, gridspec_kw={"width_ratios": [3, 1]})
    ax0, ax1 = axs

    distances, eigvals, points, labels_at_points = _get_bands_data(f_h5py_bands)

    ax0.plot(distances, eigvals[0, :, :], "ok", markersize=1)
    ax0.hlines(efermi, distances[0], distances[-1], "r", linewidth=1)
    for x in points[1:-1]:
        ax0.vlines(x, efermi + emin, efermi + emax, "k", linewidth=1)
    ax0.set_xlim(distances[0], distances[-1])
    ax0.set_xticks(points)
    ax0.set_xticklabels(labels_at_points)
    ax0.set_ylim(efermi + emin, efermi + emax)
    ax0.set_ylabel("E[eV]", fontsize=14)
    ymin, ymax = ax0.get_ylim()

    dos, energies, xmax = _get_dos_data(f_h5py_dos, ymin, ymax)

    ax1.plot(dos, energies, "-k", linewidth=1)
    ax1.hlines(efermi, 0, xmax, "r", linewidth=1)
    ax1.set_xlim(0, xmax)
    ax1.set_ylim(ymin, ymax)
    ax1.yaxis.tick_right()

    plt.rcParams["pdf.fonttype"] = 42
    plt.tight_layout()
    plt.savefig(plot_filename)
    plt.close()

    click.echo(f'Electronic band structure plot was saved in "{plot_filename}".')


def _get_bands_data(f_h5py: h5py.File):
    eigvals = f_h5py["results"]["electron_eigenvalues_kpoints_opt"]["eigenvalues"]

    reclat = get_reclat_from_vaspout(f_h5py)
    # k-points in reduced coordinates
    kpoint_coords = f_h5py["results"]["electron_eigenvalues_kpoints_opt"][
        "kpoint_coords"
    ]
    # Special point labels
    labels = [
        label.decode("utf-8")
        for label in f_h5py["input"]["kpoints_opt"]["labels_kpoints"][:]
    ]
    nk_per_seg = f_h5py["input"]["kpoints_opt"]["number_kpoints"][()]
    nk_total = len(kpoint_coords)
    k_cart = kpoint_coords @ reclat
    n_segments = nk_total // nk_per_seg
    assert n_segments * nk_per_seg == nk_total
    distances = get_distances_along_BZ_path(nk_total, n_segments, nk_per_seg, k_cart)
    points, labels_at_points = get_special_points(
        labels, distances, n_segments, nk_per_seg, nk_total
    )

    return distances, eigvals, points, labels_at_points


def _get_dos_data(f_h5py: h5py.File, ymin: float, ymax: float):
    dos = f_h5py["results"]["electron_dos_kpoints_opt"]["dos"][0, :]
    energies = f_h5py["results"]["electron_dos_kpoints_opt"]["energies"][:]
    i_min = 0
    i_max = len(energies)
    for i, val in enumerate(energies):
        if val < ymin:
            i_min = i
        if val > ymax:
            i_max = i
            break
    xmax = dos[i_min : i_max + 1].max() * 1.1
    return dos, energies, xmax



"""Implementation of velph-relax-generate."""

import copy
import pathlib
import shutil

import click
import tomli
from phonopy.interface.calculator import write_crystal_structure
from phonopy.structure.atoms import parse_cell_dict

from phelel.velph.cli.utils import (
    assert_kpoints_mesh_symmetry,
    get_scheduler_dict,
    write_incar,
    write_kpoints_line_mode,
    write_kpoints_mesh_mode,
    write_launch_script,
)


def write_input_files(toml_filename: pathlib.Path) -> None:
    """Generate VASP inputs to generate electronc band structure."""
    with open(toml_filename, "rb") as f:
        toml_dict = tomli.load(f)

    main_directory_name = "el_bands"
    main_directory = pathlib.Path(main_directory_name)
    main_directory.mkdir(parents=True, exist_ok=True)

    for calc_type in ("bands", "dos"):
        directory_name = f"{main_directory_name}/{calc_type}"
        directory = pathlib.Path(directory_name)
        directory.mkdir(parents=True, exist_ok=True)

        # POSCAR
        primitive = parse_cell_dict(toml_dict["primitive_cell"])
        write_crystal_structure(directory / "POSCAR", primitive)

        # INCAR
        if calc_type == "bands":
            incar_dict = copy.deepcopy(toml_dict["vasp"]["el_bands"]["incar"])
            for key in toml_dict["vasp"]["el_bands"]["incar"]:
                if key.lower().strip() in ("lorbit", "nedos"):
                    del incar_dict[key]
            write_incar(incar_dict, directory, cell=primitive)
        elif calc_type == "dos":
            write_incar(
                toml_dict["vasp"]["el_bands"]["incar"], directory, cell=primitive
            )

        # KPOINTS
        kpoints_dict = toml_dict["vasp"]["el_bands"]["kpoints"]
        assert_kpoints_mesh_symmetry(toml_dict, kpoints_dict, primitive)
        write_kpoints_mesh_mode(
            toml_dict["vasp"]["el_bands"]["incar"],
            directory,
            "vasp.el_bands.kpoints",
            toml_dict["vasp"]["el_bands"]["kpoints"],
        )

        # KPOINTS_OPT
        if calc_type == "bands":
            if "path" in toml_dict["vasp"]["el_bands"]["kpoints_opt"]:
                click.echo(
                    "Seek-path (https://github.com/giovannipizzi/seekpath) is used."
                )

            write_kpoints_line_mode(
                primitive,
                directory,
                "vasp.el_bands.kpoints_opt",
                toml_dict["vasp"]["el_bands"]["kpoints_opt"],
                kpoints_filename="KPOINTS_OPT",
            )
        elif calc_type == "dos":
            if "kpoints_dense" not in toml_dict["vasp"]["el_bands"]:
                raise RuntimeError(
                    "[vasp.el_bands.kpoints_dense] section is necessary "
                    "for electronic DOS calculation."
                )
            kpoints_dense_dict = toml_dict["vasp"]["el_bands"]["kpoints_dense"]
            assert_kpoints_mesh_symmetry(toml_dict, kpoints_dense_dict, primitive)
            write_kpoints_mesh_mode(
                toml_dict["vasp"]["el_bands"]["incar"],
                directory,
                "vasp.el_bands.kpoints",
                kpoints_dense_dict,
                kpoints_filename="KPOINTS_OPT",
            )

        # POTCAR
        if pathlib.Path("POTCAR").exists():
            shutil.copy2(pathlib.Path("POTCAR"), directory / "POTCAR")

        # Scheduler launch script
        if "scheduler" in toml_dict:
            scheduler_dict = get_scheduler_dict(toml_dict, "el_bands")
            write_launch_script(scheduler_dict, directory)

        click.echo(f'VASP input files were made in "{directory_name}".')



"""Utilities of main CUI script."""

import pathlib
from typing import Optional, Union

from phono3py.interface.calculator import (
    get_additional_info_to_write_fc2_supercells,
    get_additional_info_to_write_supercells,
    get_default_displacement_distance,
)
from phonopy.cui.phonopy_script import store_nac_params
from phonopy.interface.calculator import write_supercells_with_displacements

from phelel.api_phelel import Phelel


def create_phelel_supercells(
    cell_info,
    settings,
    symprec,
    interface_mode="vasp",
    load_phelel_yaml=False,
    log_level=1,
):
    """Create displacements and supercells.

    Distance unit used is that for the calculator interface.
    The default unit is Angstron.

    """
    optional_structure_info = cell_info["optional_structure_info"]
    unitcell_filename = cell_info["optional_structure_info"][0]
    phe_yml = cell_info["phonopy_yaml"]

    phelel = Phelel(
        cell_info["unitcell"],
        supercell_matrix=cell_info["supercell_matrix"],
        primitive_matrix=cell_info["primitive_matrix"],
        phonon_supercell_matrix=cell_info["phonon_supercell_matrix"],
        symprec=symprec,
        is_symmetry=settings.is_symmetry,
        calculator=interface_mode,
    )

    if log_level:
        print("")
        print(f'Unit cell was read from "{unitcell_filename}".')

    generate_phelel_supercells(
        phelel,
        interface_mode=interface_mode,
        distance=settings.displacement_distance,
        is_plusminus=settings.is_plusminus_displacement,
        is_diagonal=settings.is_diagonal_displacement,
        log_level=log_level,
    )

    if pathlib.Path("BORN").exists() or (phe_yml and phe_yml.nac_params):
        store_nac_params(
            phelel.phonon,
            settings,
            phe_yml,
            unitcell_filename,
            log_level,
            load_phonopy_yaml=load_phelel_yaml,
        )

    additional_info = get_additional_info_to_write_supercells(
        interface_mode, phelel.supercell_matrix
    )
    write_supercells_with_displacements(
        interface_mode,
        phelel.supercell,
        phelel.supercells_with_displacements,
        optional_structure_info=optional_structure_info,
        additional_info=additional_info,
    )

    if phelel.phonon is not None:
        additional_info = get_additional_info_to_write_fc2_supercells(
            interface_mode, phelel.phonon_supercell_matrix, suffix="PH"
        )
        write_supercells_with_displacements(
            phelel.phonon.calculator,
            phelel.phonon.supercell,
            phelel.phonon.supercells_with_displacements,
            optional_structure_info=optional_structure_info,
            additional_info=additional_info,
        )

    return phelel


def generate_phelel_supercells(
    phelel: Phelel,
    interface_mode: str = "vasp",
    distance: Optional[float] = None,
    is_plusminus: Union[str, bool] = "auto",
    is_diagonal: bool = True,
    log_level: int = 0,
):
    """Generate phelel supercells."""
    if distance is None:
        _distance = get_default_displacement_distance(interface_mode)
    else:
        _distance = distance

    phelel.generate_displacements(
        distance=_distance, is_plusminus=is_plusminus, is_diagonal=is_diagonal
    )

    if log_level:
        print("Displacement distance: %s" % _distance)
        print("Number of displacements: %d" % len(phelel.supercells_with_displacements))

    if phelel.phonon is not None:
        phelel.phonon.generate_displacements(
            distance=distance, is_plusminus=is_plusminus, is_diagonal=is_diagonal
        )
        if log_level:
            print(
                "Number of displacements for phonon: %d"
                % len(phelel.phonon.supercells_with_displacements)
            )



"""Implementation of velph-transport-plot-eigenvalues."""

from __future__ import annotations

from typing import Optional

import click
import h5py
import matplotlib.pyplot as plt
import numpy as np
from phonopy.structure.brillouin_zone import get_qpoints_in_Brillouin_zone
from phonopy.structure.cells import get_reduced_bases
from phonopy.units import Kb
from scipy.spatial import Voronoi

from phelel.velph.cli.utils import get_symmetry_dataset
from phelel.velph.utils.vasp import read_crystal_structure_from_h5


def fermi_dirac_distribution(energy: np.ndarray, temperature: float) -> np.ndarray:
    """Calculate the Fermi-Dirac distribution.

    energy in eV measured from chemical potential
    temperature in K

    """
    de = energy / (Kb * temperature)
    de = np.where(de < 100, de, 100.0)  # To avoid overflow
    de = np.where(de > -100, de, -100.0)  # To avoid underflow
    return 1 / (1 + np.exp(de))


def plot_eigenvalues(
    f_h5py: h5py.File,
    tid: Optional[int] = None,
    temperature: Optional[float] = None,
    cutoff_occupancy: float = 1e-2,
    mu: Optional[float] = None,
    time_reversal: bool = True,
):
    """Show eigenvalues, occupation, k-points and Fermi-Dirac distribution.

    Parameters
    ----------
    f_h5py : h5py.File
        vaspout.h5 file object.
    temperature : float
        Temperature for Fermi-Dirac distribution in K.
    cutoff_occupancy : float
        Cutoff for the occupancy to show eigenvalues in eV. Eigenvalus with
        occupances in interval [cutoff_occupancy, 1 - cutoff_occupancy] is
        shown.
    mu : float or None
        Chemical potential in eV. If None, the Fermi energy.

    """
    cell = read_crystal_structure_from_h5(f_h5py, "results/positions")
    sym_dataset = get_symmetry_dataset(cell)
    rotations = [r.T for r in sym_dataset.rotations]

    if tid is not None:
        transport = f_h5py["results/electron_phonon/electrons/transport_1"]
        _temperature = transport["temps"][tid - 1]
        _mu = transport["mu"][tid - 1]
    else:
        if temperature is None:
            _temperature = 300
        else:
            _temperature = temperature
        if mu is None:
            _mu = f_h5py["results/electron_phonon/electrons/dos/efermi"][()]
        else:
            _mu = mu

    click.echo(f"mu: {_mu:.6f} eV")
    click.echo(f"temperature: {_temperature:.6f} K")

    if time_reversal:
        has_inversion = False
        for r in rotations:
            if (r == -np.eye(3, dtype=int)).all():
                has_inversion = True
                break
        if not has_inversion:
            rotations += [-r for r in rotations]

    dir_eigenvalues = "results/electron_phonon/electrons/eigenvalues"
    eigenvals = f_h5py[f"{dir_eigenvalues}/eigenvalues"][:] - _mu
    # weights = f_h5py[f"{dir_eigenvalues}/fermiweights"][:]
    kpoints = f_h5py[f"{dir_eigenvalues}/kpoint_coords"][:]
    ind = np.unravel_index(np.argsort(-eigenvals, axis=None), eigenvals.shape)
    weights = fermi_dirac_distribution(eigenvals, _temperature)

    all_kpoints = []
    all_weights = []
    all_eigenvals = []
    for i, (e, wt) in enumerate(zip(eigenvals[ind], weights[ind])):
        k = kpoints[ind[1][i]]
        if wt < cutoff_occupancy or wt > 1 - cutoff_occupancy:
            continue
        all_kpoints += list(rotations @ k.T)
        all_weights += [wt] * len(rotations)
        all_eigenvals += [e] * len(rotations)

    if not all_kpoints:
        click.echo("No eigenvalues to plot.")
        return

    all_kpoints = np.array(all_kpoints)
    all_kpoints -= np.rint(all_kpoints)
    all_kpoints = get_qpoints_in_Brillouin_zone(
        np.linalg.inv(cell.cell), all_kpoints, only_unique=True
    )
    all_kpoints = all_kpoints @ np.linalg.inv(cell.cell).T
    all_weights = np.array(all_weights)
    all_eigenvals = np.array(all_eigenvals)

    with open("bz.dat", "w") as w:
        for i, (e, wt, rk) in enumerate(zip(all_eigenvals, all_weights, all_kpoints)):
            print(
                f"{i + 1} {e:.6f} {wt:.6f} [{rk[0]:.6f} {rk[1]:.6f} {rk[2]:.6f}]",
                file=w,
            )

    _plot_eigenvalues_in_BZ(
        all_kpoints,
        all_weights,
        np.linalg.inv(cell.cell),
        title=f"mu={_mu:.6f} eV, temperature={_temperature:.1f} K",
    )


def _plot_eigenvalues_in_BZ(
    data: np.ndarray,
    weights: np.ndarray,
    bz_lattice: np.ndarray,
    title: Optional[str] = None,
):
    """Plot kpoints in Brillouin zone."""
    ax = _get_ax_3D()
    point_sizes = (1 - np.abs(weights)) * 5
    scatter = ax.scatter(
        data[:, 0],
        data[:, 1],
        data[:, 2],
        s=point_sizes,
        c=weights,
        cmap="viridis",
    )
    _plot_Brillouin_zone(bz_lattice, ax)
    cbar = plt.colorbar(scatter, ax=ax)
    cbar.set_label("Occupancy")
    cbar.ax.yaxis.label.set_rotation(-90)
    cbar.ax.yaxis.labelpad = 20

    if title:
        ax.set_title(title)
    plt.show()


def _get_ax_3D() -> plt.Axes:
    """Get 3D axis."""
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection="3d")
    ax.grid(False)
    ax.set_box_aspect([1, 1, 1])
    return ax


def _plot_Brillouin_zone(bz_lattice: np.ndarray, ax: plt.Axes):
    """Plot Brillouin zone.

    Parameters
    ----------
    bz_lattice : np.ndarray
        Reciprocal basis vectors in column vectors.

    """
    bz_red_lattice = get_reduced_bases(bz_lattice.T).T
    points = np.dot(np.array(list(np.ndindex(3, 3, 3))) - [1, 1, 1], bz_red_lattice.T)
    vor = Voronoi(points)

    distances_to_origin = np.linalg.norm(points, axis=1)
    origin_point_index = np.argmin(distances_to_origin)

    for ridge_points, ridge_cells in zip(vor.ridge_vertices, vor.ridge_points):
        if -1 in ridge_points:
            continue

        if origin_point_index in ridge_cells:
            # The first point is appended to the end to close the loop.
            _ridge_points = ridge_points[:] + [ridge_points[0]]
            ridge_vertices = np.array([vor.vertices[i] for i in _ridge_points])
            ax.plot(
                ridge_vertices[:, 0],
                ridge_vertices[:, 1],
                ridge_vertices[:, 2],
                ":",
                linewidth=0.5,
            )



"""Convenient routines to show detailed data."""

import io
import os

import h5py
import numpy as np
from phonopy.interface.vasp import VasprunxmlExpat, get_vasp_structure_lines, read_vasp

from phelel.interface.vasp.file_IO import (
    get_CHGCAR,
    read_dprojectors,
    read_eigenvalues,
    read_inwap_yaml,
    read_local_potential,
    read_PAW_Dij_qij,
    read_qtot,
    read_waves,
)
from phelel.interface.vasp.procar import QTOT, CoreCharge, Procar


class DijQij:
    """Container of Dij and qij.

    Attributes
    ----------
    Dij : ndarray
        PAW strength
        dtype=complex128
        shape=(spin, atom, lm, l'm')
    qij : ndarray
        <phi_i|phi_j>-<phi_i~|phi_j~>
        dtype=complex128
        shape=(spin, atom, lm, l'm')

    """

    def __init__(self, Dij, qij):
        """Init method.

        Note
        ----
        The order of lm is:
            (l=0,m=0), (l=1,m=-1), (l=1,m=0), (l=1,m=1), (l=2,m=-2),
            (l=2,m=-1), (l=2,m=0), ...

        Parameters
        ----------
        Dij : ndarray
            PAW strength
            dtype=complex128
            shape=(spin, atom, lm, l'm')
        qij : ndarray
            <phi_i|phi_j>-<phi_i~|phi_j~>
            dtype=complex128
            shape=(spin, atom, lm, l'm')

        """
        self.Dij = Dij
        self.qij = qij

    def __str__(self):
        """Generate qij and Dij as text."""
        # qij are real and Dij are expected to be real.
        text = self._str_format(self.Dij.real, "Dij real part")
        # text += self._str_format(self.Dij.imag, "Dij imaginary part")
        text += self._str_format(self.qij.real, "qij real part")
        # text += self._str_format(self.qij.imag, "qij imaginary part")
        return text

    def _str_format(self, val, comment=""):
        text = ""
        for i, dij_spin in enumerate(val):
            for j, dij_atom in enumerate(dij_spin):
                text += "Spin %d, Atom %d, %s (%d, %d)\n" % (
                    (i + 1, j + 1) + (comment,) + dij_atom.shape
                )
                for dij_l in dij_atom:
                    # if (abs(dij_l) > 1e-4).any():
                    #     text += " ".join(["%8.4f" % x for x in dij_l]) + "\n"
                    # else:
                    #     text += " ".join(["%8.1e" % x for x in dij_l]) + "\n"
                    text += " ".join(["%8.1e" % x for x in dij_l]) + "\n"
        return text


class LocalPotential:
    """Container of Local potential."""

    def __init__(self, V_loc):
        """Init method.

        Parameters
        ----------
        V_loc : ndarray
            V_loc stores local potential.
            dtype=complex128
            shape=(ncdij, nz, ny, nx)

            ncdij=1 without spin
            ncdij=2 with collinear spin (up and down)
            ncdij=4 with non-collinear spin (2x2 localpotential)

        """
        self._V_loc = V_loc

    @property
    def V(self):
        """Return local potential."""
        return self._V_loc

    def write_locpot(self, header, spin_polarized=False, filename="locpot.dat"):
        """Write local potential to locpot.dat."""
        text = get_CHGCAR(self._V_loc[0].real, header)
        with open(filename, "w") as w:
            w.write(text)

    def __str__(self):
        """Return array shape of local potential as text."""
        text = "Local potential (ncdij, nz, ny, nx) = "
        text += str(self._V_loc.shape) + "\n"
        # text += str(self._V_loc)
        return text


class WaveFunction:
    """Container of pseudo wave functions."""

    def __init__(self, waves, weights=None, occupancy=None):
        """Init method.

        Parameters
        ----------
        waves : ndarray
            wavefunctions
            shape=(nkpts, nbtot, ispin, nrspinors, nz, ny, nx)
            dtype=complex128
        weights : list of float values
            weights of k-points. This is not given by wap.
            shape=(nkpts,)
        occupancy : list of lists of float values
            Electron occupations at bands and k-points. This is not given by
            wap.
            shape=(nkpts, nbtot)

        """
        self._waves = waves
        self._weights = weights
        self._occupancy = occupancy

    def write_charge(self, header, spin_polarized=False, filename="charge.dat"):
        """Write sum of squared orbitals to charge.dat."""
        charge = self._sum_charge(spin_polarized)
        text = get_CHGCAR(charge, header)
        with open(filename, "w") as w:
            w.write(text)
        return charge.sum() / np.prod(charge.shape)

    def _square_sum(self, ikpt=0, iband=0, ispin=0, ispinor=0):
        wave = self._waves[ikpt, iband, ispin, ispinor]
        return (wave.real**2 + wave.imag**2).sum()

    def _sum_charge(self, spin_polarized):
        """Calculate pseudo charge density.

        Note that wave functions are not normalized by themselves due to overlap
        matrix. Charge density out of PAW spheres is expected to agree with
        CHGCAR except for mesh density.

        """
        charge = self._waves.real**2 + self._waves.imag**2
        sum_charge = charge.sum(axis=2).sum(axis=2)
        if spin_polarized:
            occupancy = self._occupancy
        else:
            occupancy = np.array(self._occupancy) * 2

        for i, w in enumerate(self._weights):
            for j, occ in enumerate(occupancy[i]):
                sum_charge[i, j] *= w * occ
        sum_charge = sum_charge.sum(axis=0).sum(axis=0)
        return sum_charge

    def __str__(self):
        """Return total sum of squared orbitals on grid as text."""
        text = "Waves (nkpts, nbtot, ispin, nrspinors, nz, ny, nx) = "
        text += str(self._waves.shape)
        text += "\n"
        # text += str(self._waves)
        for ikpt, iband in list(np.ndindex(self._waves.shape[:2])):
            text += "sum_i|pseudo-wave_i|^2 at k-point %d, band %d: %f\n" % (
                ikpt + 1,
                iband + 1,
                self._square_sum(ikpt=ikpt, iband=iband),
            )
        return text


class VaspShowData:
    """Show el-ph related data generated by VASP."""

    def __init__(self):
        """Init method."""
        self.parse()

    def parse(self):
        """Parse files."""
        self._inwap = read_inwap_yaml(filename="inwap.yaml")
        self._Dij = read_PAW_Dij_qij(self._inwap, "PAW-STRENGTH.bin")
        self._qij = read_PAW_Dij_qij(self._inwap, "PAW-OVERLAP.bin")
        self._qtot = read_qtot(self._inwap, filename="QTOT.bin")
        self._V_loc = read_local_potential(self._inwap, filename="LOCAL-POTENTIAL.bin")

        if os.path.exists("EIGENVALUE.bin"):
            self._eigvals = read_eigenvalues(self._inwap, filename="EIGENVALUE.bin")
        else:
            self._eigvals = None
        if os.path.exists("WAVES.bin"):
            self._waves = read_waves(self._inwap, filename="WAVES.bin")
        else:
            self._waves = None
        if os.path.exists("DPROJECTORS.bin"):
            self._proj = read_dprojectors(self._inwap, filename="DPROJECTORS.bin")
            filename = "dprojects.hdf5"
            w = h5py.File(filename, "w")
            w.create_dataset("cproj", data=self._proj)
            w.close()
        else:
            self._proj = None

        self._cell_el = read_vasp("POSCAR")
        self._parse_vasprun_xml()

    def show(self):
        """Show data."""
        if self._waves is not None:
            print(self.waves)
        print(self.dij)
        print(self.V_loc)
        print(self.Q_tot)
        if self._proj is not None:
            print(self.procar)
            print(self.core_charge)
        print(self.eigenvalues)

    def chgcar(self):
        """Return total charge and write orbital squared to a file."""
        header = "\n".join(get_vasp_structure_lines(self._cell_el))
        n_charge = self.waves.write_charge(header)

        msg = (
            "VASP CHGCAR like data is generated from wave function. "
            "k-point symmetry is not expanded in phelel, therefore "
            "to obtain very similar results to CHGCAR, ISYM=0 has to "
            "be specified."
        )
        import textwrap

        text = textwrap.fill(msg, width=70) + "\n\n"
        text += 'The result is written into "charge.dat".\n'
        text += "Total number of electrons = %f" % n_charge
        return text

    def locpot(self):
        """Write local potential to locpot.dat."""
        header = "\n".join(get_vasp_structure_lines(self._cell_el))
        self.V_loc.write_locpot(header)

        msg = (
            'The local potential is written in "locpot.dat" in VASP '
            "LOCPOT-like format."
        )
        import textwrap

        text = textwrap.fill(msg, width=70)
        return text

    @property
    def dij(self):
        """Return DijQij class instance."""
        dij = DijQij(self._Dij, self._qij)
        return dij

    @property
    def V_loc(self):
        """Return LocalPotential class instance."""
        V_loc = LocalPotential(self._V_loc)
        return V_loc

    @property
    def eigenvalues(self):
        """Return Eigenvalue class instance."""
        e = Eigenvalue(self._eigvals)
        return e

    @property
    def waves(self):
        """Return WaveFunction class instance."""
        occupancy = []
        for eig_k in self._xml_eigvals[0]:  # spin 1
            occupancy.append([eig_b[1] for eig_b in eig_k])
        waves = WaveFunction(self._waves, weights=self._k_weights, occupancy=occupancy)
        return waves

    @property
    def Q_tot(self):
        """Return QTOT class instance."""
        qtot = QTOT(self._qtot)
        return qtot

    @property
    def procar(self):
        """Return Procar class instance."""
        proj = Procar(self._proj, self._inwap["lm_orbitals"], self._qtot)
        return proj

    @property
    def core_charge(self):
        """Return CoreCharge class instance."""
        rho = CoreCharge(self._proj, self._qij)
        return rho

    def _parse_vasprun_xml(self, filename="vasprun.xml"):
        with io.open(filename, "rb") as f:
            vxml = VasprunxmlExpat(f)
            if vxml.parse():
                self._xml_eigvals = vxml.get_eigenvalues()
                self._k_weights = vxml.get_k_weights()
            else:
                print("Parsing vasprun.xml failed.")


class Eigenvalue:
    """Container of eigenvalues."""

    def __init__(self, e):
        """Init method.

        Parameters
        ----------
        e : ndarray
            Eigenvalues
            dtype='double'
            shape=(nbtot, nkpts, ispin)

        """
        self._e = e

    def __str__(self):
        """Generate eigenvalues information as text."""
        text = "Eigenvalues (nbtot, nkpts, ispin) = "
        text += str(self._e.shape) + "\n"
        for i in range(self._e.shape[2]):
            for j, e_k in enumerate(self._e[:, :, i].T):
                text += "ISPIN: %d/%d, k-point: %d/%d\n" % (
                    i + 1,
                    self._e.shape[2],
                    j + 1,
                    self._e.shape[1],
                )
                text += "%5s %16s\n" % ("Band", "Eigenvalue")
                for k, e_b in enumerate(e_k):
                    text += "%5d %16.6f\n" % (k + 1, e_b)
                text += "\n"
        return text



"""Utilities for cli."""

from __future__ import annotations

import copy
import os
import pathlib
import xml.parsers.expat
from dataclasses import asdict, dataclass
from enum import Enum
from typing import Literal, Optional, Union

try:
    from spglib import SpglibDataset
except ImportError:
    from types import SimpleNamespace as SpglibDataset

import click
import h5py
import numpy as np
import spglib
from phono3py.phonon.grid import BZGrid
from phonopy.interface.vasp import VasprunxmlExpat, sort_positions_by_symbols
from phonopy.structure.atoms import PhonopyAtoms, atom_data, parse_cell_dict
from phonopy.structure.cells import (
    get_primitive,
    get_primitive_matrix_by_centring,
    get_reduced_bases,
)
from phonopy.structure.symmetry import symmetrize_borns_and_epsilon
from phonopy.units import Bohr, Hartree
from phonopy.utils import get_dot_access_dataset

from phelel.velph.utils.scheduler import (
    get_custom_schedular_script,
    get_sge_scheduler_script,
    get_slurm_scheduler_script,
)
from phelel.velph.utils.vasp import VaspIncar, VaspKpoints


class CellChoice(Enum):
    """Cell choice for specific calculations."""

    UNSPECIFIED = "unspecified"
    UNITCELL = "unitcell"
    PRIMITIVE = "primitive"


class PrimitiveCellChoice(Enum):
    """Primitive cell choice."""

    STANDARDIZED = "standardized"
    REDUCED = "reduced"


@dataclass(frozen=True)
class DefaultCellChoices:
    """Default cell choices."""

    nac: CellChoice = CellChoice.PRIMITIVE
    relax: CellChoice = CellChoice.UNITCELL


@dataclass(frozen=True)
class VelphInitParams:
    """Basic init parameters of velph.

    amplitude, diagonal, plusminus, phelel_nosym can be parameters for [phelel].
    These are treated specially.

    Except for amplitude, diagonal, plusminus, phelel_nosym, None is equivalent
    to its default value.

    """

    amplitude: Optional[float] = 0.03
    cell_for_nac: Optional[CellChoice] = CellChoice.UNSPECIFIED
    cell_for_relax: Optional[CellChoice] = CellChoice.UNSPECIFIED
    find_primitive: Optional[bool] = True
    diagonal: Optional[bool] = True
    plusminus: Optional[bool] = True
    kspacing: Optional[float] = 0.1
    kspacing_dense: Optional[float] = 0.05
    magmom: Optional[str] = None
    max_num_atoms: Optional[int] = None
    phonon_max_num_atoms: Optional[int] = None
    phelel_nosym: Optional[bool] = False
    primitive_cell_choice: Optional[PrimitiveCellChoice] = (
        PrimitiveCellChoice.STANDARDIZED
    )
    symmetrize_cell: Optional[bool] = False
    tolerance: Optional[float] = 1e-5
    use_grg: Optional[bool] = False


@dataclass(frozen=True)
class VelphFilePaths:
    """File paths or pointers used in velph-init."""

    cell_filepath: pathlib.Path
    velph_template_filepath: Optional[pathlib.Path] = None


def write_incar(
    toml_incar_dict: dict,
    directory: pathlib.Path,
    cell: Optional[PhonopyAtoms] = None,
    incar_filename: Union[str, bytes, os.PathLike] = "INCAR",
) -> None:
    """Write INCAR file."""
    incar_dict = copy.deepcopy(toml_incar_dict)
    if cell.magnetic_moments is not None:
        incar_dict["magmom"] = cell.magnetic_moments.tolist()
    VaspIncar.write(directory / incar_filename, incar_dict)


def write_kpoints_mesh_mode(
    toml_incar_dict: dict,
    directory: pathlib.Path,
    tag: str,
    toml_kpoints_dict: dict,
    kpoints_filename="KPOINTS",
    kspacing_name="kspacing",
) -> None:
    """Write KPOINTS file in mesh mode."""
    if toml_incar_dict.get(kspacing_name) is None:
        try:
            VaspKpoints.write_mesh_mode(directory / kpoints_filename, toml_kpoints_dict)
        except KeyError:
            click.echo(
                f'Invalid setting of [{tag}]. "{kpoints_filename}" was not made.'
            )
    else:
        click.echo(
            f'"{kpoints_filename}" was not made because of '
            f'"{kspacing_name}" tag in INCAR setting.'
        )


def write_kpoints_line_mode(
    cell: PhonopyAtoms,
    directory: pathlib.Path,
    tag: str,
    toml_kpoints_dict: dict,
    kpoints_filename: Union[pathlib.Path, str] = "KPOINTS",
) -> None:
    """Write KPOINTS file in line mode."""
    VaspKpoints.write_line_mode(directory / kpoints_filename, cell, toml_kpoints_dict)


def write_launch_script(
    toml_scheduler_dict: dict, directory: pathlib.Path, job_id: Optional[str] = None
) -> None:
    """Write scheduler launch script."""
    sched_string = None
    if toml_scheduler_dict["scheduler_name"] == "sge":
        sched_string = get_sge_scheduler_script(toml_scheduler_dict, job_id=job_id)
    elif toml_scheduler_dict["scheduler_name"] == "slurm":
        sched_string = get_slurm_scheduler_script(toml_scheduler_dict, job_id=job_id)
    elif toml_scheduler_dict["scheduler_name"] == "custom":
        if "custom_template" in toml_scheduler_dict:
            raise RuntimeError(
                'Key "custom_template" is obsoleted. Use "scheduler_template".'
            )

        if "scheduler_template" not in toml_scheduler_dict:
            click.echo(
                '"scheduler_template" has to be specified in scheduler setting.',
                err=True,
            )
        else:
            sched_string = get_custom_schedular_script(
                toml_scheduler_dict["scheduler_template"],
                toml_scheduler_dict,
                job_id=job_id,
            )

    if sched_string:
        with open(directory / "_job.sh", "w") as w:
            w.write(sched_string)


def get_scheduler_dict(toml_dict: dict, calc_type: str) -> dict:
    """Collect and return scheduler information.

    This function extracts scheduler dict from ``toml_dict``.

    Scheduler information is written in [scheduler] section. This information
    can be overwritten in ``each calc_type`` by writing parameters in
    [vasp.calc_type.scheduler] section.

    Parameters
    ----------
    toml_dict : dict
        Dictionary after parsing velph.toml.
    calc_type : str
        This can be written in either way "supercell" or "supercell.phonon".
        This string is splitted by period (".") and the list of strings is used
        as keys of ``toml_dict["vasp"]["key1"]["key2"]``.

    """
    scheduler_dict = copy.deepcopy(toml_dict["scheduler"])
    calc_type_keys = calc_type.split(".")
    tmp_dict = toml_dict["vasp"]
    for key in calc_type_keys:
        tmp_dict = tmp_dict[key]
    if "scheduler" in tmp_dict:
        scheduler_dict.update(tmp_dict["scheduler"])
    return scheduler_dict


def assert_kpoints_mesh_symmetry(
    toml_dict: dict, kpoints_dict: dict, primitive: PhonopyAtoms
):
    """Check if mesh grid respects crystallographic point group or not."""
    if "kspacing" in kpoints_dict:
        symmetry_dataset = kspacing_to_mesh(kpoints_dict, primitive)
        if "symmetry" in toml_dict and "spacegroup_type" in toml_dict["symmetry"]:
            assert (
                symmetry_dataset.international
                == toml_dict["symmetry"]["spacegroup_type"]
            )


def choose_cell_in_dict(
    toml_dict: dict,
    toml_filename: str,
    calc_type: Literal["relax", "nac"],
) -> Optional[PhonopyAtoms]:
    """Return unit cell, primitive cell, or Niggli reduced cell from toml_dict.

    Unit cell and primitive cell have to exist in velph.toml. But Niggli reduced
    cell is optional.

    """
    if "cell" in toml_dict["vasp"][calc_type]:
        if "unitcell" in toml_dict["vasp"][calc_type]["cell"]:
            cell = parse_cell_dict(toml_dict["unitcell"])
        elif "primitive" in toml_dict["vasp"][calc_type]["cell"]:
            cell = parse_cell_dict(toml_dict["primitive_cell"])
        else:
            msg = (
                f"[vasp.{calc_type}] cell in {toml_filename} has to be either "
                '"unitcell" or "primitive_cell"'
            )
            click.echo(msg, err=True)
            return None
    else:
        if asdict(DefaultCellChoices())[calc_type] is CellChoice.PRIMITIVE:
            cell = parse_cell_dict(toml_dict["primitive_cell"])
            click.echo(f"Primitive cell was used for {calc_type}.")
        elif asdict(DefaultCellChoices())[calc_type] is CellChoice.UNITCELL:
            cell = parse_cell_dict(toml_dict["unitcell"])
            click.echo(f"Unitcell was used for {calc_type}.")
        else:
            raise RuntimeError("This should not happen.")

    return cell


def get_reclat_from_vaspout(fp_vaspout: h5py.File):
    """Return reciprocal space basis vectors.

    Returns
    -------
    reclat : np.ndarray
        Reciprocal basis vectors in row vectors.
        shape=(3, 3)

    """
    # Basis vectors in direct space in column vectors
    lattice = np.transpose(
        fp_vaspout["input"]["poscar"]["lattice_vectors"][:]
        * fp_vaspout["input"]["poscar"]["scale"][()]
    )
    # Basis vectors in reciprocal space in row vectors
    reclat = 2 * np.pi * np.linalg.inv(lattice)
    return reclat


def get_distances_along_BZ_path(nk_total, n_segments, nk_per_seg, k_cart):
    """Measure distances of points from origin along paths.

    Returns
    -------
    distances : np.ndarray
        Distances of points from origin along BZ-paths.
        shape=(nk_total,)

    """
    distances = np.zeros(nk_total)
    count = 0
    for _ in range(n_segments):
        for i_pts in range(nk_per_seg):
            # Treatment of jump between equivalent points on BZ boundary
            if i_pts == 0:
                delta_dist = 0
            else:
                delta_dist = np.linalg.norm(k_cart[count, :] - k_cart[count - 1, :])
            distances[count] = distances[count - 1] + delta_dist
            count += 1
    return distances


def get_special_points(
    labels: list[str],
    distances: list[float],
    n_segments: int,
    nk_per_seg: int,
    nk_total: int,
) -> tuple[list, list]:
    """Plot special points at vertical lines and labels."""
    # Left most
    points = []
    labels_at_points = []

    labels_at_points.append(labels[0][0])
    points.append(distances[0])

    count = 0
    for i_seg in range(n_segments):
        for _ in range(nk_per_seg):
            count += 1
        if count != nk_total:
            points.append(distances[count])
            if labels[i_seg * 2 + 1] == labels[i_seg * 2 + 2]:
                labels_at_points.append(labels[i_seg * 2 + 1][0])
            else:
                labels_at_points.append(
                    f"{labels[i_seg * 2 + 1][0]}|{labels[i_seg * 2 + 2][0]}"
                )

    labels_at_points.append(labels[-1][0])
    points.append(distances[-1])

    return points, labels_at_points


def get_symmetry_dataset(cell: PhonopyAtoms, tolerance: float = 1e-5) -> SpglibDataset:
    """Return spglib symmetry dataset."""
    dataset = get_dot_access_dataset(
        spglib.get_symmetry_dataset(cell.totuple(), symprec=tolerance)
    )
    return dataset


def get_primitive_cell(
    cell: PhonopyAtoms, sym_dataset: SpglibDataset, tolerance: float = 1e-5
) -> tuple[PhonopyAtoms, np.ndarray]:
    """Return primitive cell and transformation matrix.

    This primitive cell is generated from the input cell without
    rigid rotation in contrast to `_get_standardized_unitcell`.

    """
    tmat = sym_dataset.transformation_matrix
    centring = sym_dataset.international[0]
    pmat = get_primitive_matrix_by_centring(centring)
    total_tmat = np.array(np.dot(np.linalg.inv(tmat), pmat), dtype="double", order="C")

    return (
        get_primitive(cell, primitive_matrix=total_tmat, symprec=tolerance),
        total_tmat,
    )


def get_reduced_cell(
    cell: PhonopyAtoms, method="niggli", tolerance: float = 1e-5
) -> PhonopyAtoms:
    """Return a reduced cell of input cell."""
    reduced_lattice = get_reduced_bases(cell.cell, method=method, tolerance=tolerance)
    reduced_positions = cell.scaled_positions @ (
        cell.cell @ np.linalg.inv(reduced_lattice)
    )
    reduced_positions = reduced_positions - np.rint(reduced_positions)
    reduced_positions[:, :] += np.where(reduced_positions < 0, 1, 0)
    reduced_cell = cell.copy()
    reduced_cell.cell = reduced_lattice
    reduced_cell.scaled_positions = reduced_positions
    return reduced_cell


def generate_standardized_cells(
    sym_dataset: SpglibDataset,
    tolerance: float = 1e-5,
) -> tuple[PhonopyAtoms, PhonopyAtoms, np.ndarray]:
    """Return standardized unit cell and primitive cell."""
    click.echo(
        "Crystal structure was standardized based on space-group-type "
        f"{sym_dataset.international}."
    )
    convcell = get_standardized_unitcell(sym_dataset)
    centring = sym_dataset.international[0]
    pmat = get_primitive_matrix_by_centring(centring)
    if centring == "P":
        primitive = convcell
    else:
        primitive = get_primitive(convcell, primitive_matrix=pmat, symprec=tolerance)

    return convcell, primitive, pmat


def get_standardized_unitcell(dataset: SpglibDataset) -> PhonopyAtoms:
    """Return conventional unit cell.

    This conventional unit cell can include rigid rotation with respect to
    input unit cell for which symmetry was analized.

    Parameters
    ----------
    cell : PhonopyAtoms
        Input cell.
    dataset : SpgliDataset
        Symmetry dataset of spglib.

    Returns
    -------
    PhonopyAtoms
        Convetional unit cell.

    """
    std_positions = dataset.std_positions
    std_types = dataset.std_types
    _, _, _, perm = sort_positions_by_symbols(std_types, std_positions)
    convcell = PhonopyAtoms(
        cell=dataset.std_lattice,
        scaled_positions=std_positions[perm],
        symbols=[atom_data[n][1] for n in std_types[perm]],
    )
    return convcell


def get_num_digits(sequence, min_length=3):
    """Return number of digits of sequence."""
    nd = len(str(len(sequence)))
    if nd < min_length:
        nd = min_length
    return nd


def kspacing_to_mesh(
    kpoints_dict: dict, unitcell: PhonopyAtoms, use_grg: bool = False
) -> SpglibDataset:
    """Update kpoints_dict by mesh corresponding to kspacing.

    Parameters
    ----------
    kpoints_dict : dict
        E.g., {"kspacing": 0.5}
    lattice : array_like
        Basis vectors in row vectors.
        shape=(3, 3)

    Returns
    -------
    SpglibDataset
        Symmetry dataset of spglib.

    """
    kspacing = kpoints_dict["kspacing"]
    symmetry_dataset = get_symmetry_dataset(unitcell)
    bzgrid = BZGrid(
        2 * np.pi / kspacing,
        lattice=unitcell.cell,
        symmetry_dataset=symmetry_dataset,
        use_grg=use_grg,
    )
    if bzgrid.grid_matrix is None:
        kpoints_dict["mesh"] = bzgrid.D_diag.tolist()
    else:
        kpoints_dict["mesh"] = bzgrid.grid_matrix.tolist()
    return symmetry_dataset


def check_fft(toml_filename: str, calculation_name: str) -> None:
    """Show [NGX, NGY, NGZ] in vasprun.xml."""
    vasprun_path = pathlib.Path(calculation_name) / "vasprun.xml"
    if vasprun_path.is_file():
        with open(vasprun_path, "rb") as f:
            vasprun = VasprunxmlExpat(f)
            vasprun.parse()
            click.echo(
                f'Modify [phelel] section in "{toml_filename}" '
                f'as "fft_mesh = {vasprun.fft_grid}"'
            )
    else:
        click.echo(f'"{vasprun_path}" not found.')
        click.echo("For estimating FFT mesh numbers, prepare dry-run by")
        click.echo(f"velph elph generate -d -c {calculation_name}")


def get_nac_params(
    toml_dict: dict,
    vasprun_path: pathlib.Path,
    primitive: Optional[PhonopyAtoms],
    convcell: PhonopyAtoms,
    is_symmetry: bool,
    symprec: float = 1e-5,
) -> Optional[dict]:
    """Collect NAC parameters from vasprun.xml and return them."""
    with open(vasprun_path, "rb") as f:
        try:
            vasprun = VasprunxmlExpat(f)
            vasprun.parse()
        except xml.parsers.expat.ExpatError:
            click.echo(f'Parsing "{vasprun_path}" failed.')
            return None

    nac_cell = convcell
    try:
        if "primitive" in toml_dict["vasp"]["nac"]["cell"]:
            nac_cell = primitive
    except KeyError:
        pass

    borns_, epsilon_ = symmetrize_borns_and_epsilon(
        vasprun.born,
        vasprun.epsilon,
        nac_cell,
        primitive=primitive,
        symprec=symprec,
        is_symmetry=is_symmetry,
    )

    nac_params = {
        "born": borns_,
        "factor": Hartree * Bohr,
        "dielectric": epsilon_,
    }
    return nac_params



"""Implementation of velph-transport-plot-selfenergy."""

from __future__ import annotations

import click
import h5py
import matplotlib.pyplot as plt


def plot_selfenergy(f_h5py: h5py.File, plot_filename: str, save_plot: bool = False):
    """Plot imaginary part of self-energies.

    Number of "self_energy_*" is

    (N(delta) * N(nbands_sum_array) * N(selfen_approx))
      * N(ncarrier_per_cell) * N(ncarrier_den) * N(mu)

    sefeln_approx includes
    - scattering_approximation (CRTA, ERTA, MRTA, MRTA2)
    - static_approximation (True or False)

    """
    selfens = {}
    for key in f_h5py["results"]["electron_phonon"]["electrons"]:
        if "self_energy_" in key:
            selfens[int(key.split("_")[2])] = f_h5py["results"]["electron_phonon"][
                "electrons"
            ][key]

    if len(selfens) == 1:
        fig, axs = plt.subplots(1, 1, figsize=(4, 4))
    else:
        nrows = len(selfens) // 2
        fig, axs = plt.subplots(nrows, 2, figsize=(8, 4 * nrows), squeeze=True)

    for i in range(len(selfens)):
        selfen = selfens[i + 1]
        _show(selfen, i + 1)
        _plot(axs[i], selfen)

    plt.tight_layout()
    if save_plot:
        plt.rcParams["pdf.fonttype"] = 42
        plt.savefig(plot_filename)
        click.echo(f'Transport plot was saved in "{plot_filename}".')
    else:
        plt.show()
    plt.close()


def _plot(ax, selfen):
    for i_nw in range(selfen["nw"][()]):
        for i_temp, _ in enumerate(selfen["temps"]):
            ax.plot(
                selfen["energies"][:, i_nw],
                selfen["selfen_fan"][:, i_nw, i_temp, 1],
                ".",
            )


def _show(selfen: h5py.Group, index: int):
    """Show self-energy properties.

    ['band_start', 'band_stop', 'bks_idx', 'carrier_per_cell',
    'carrier_per_cell0', 'delta', 'efermi', 'energies', 'enwin', 'nbands',
    'nbands_sum', 'nw', 'scattering_approximation', 'select_energy_window',
    'selfen_dw', 'selfen_fan', 'static', 'tetrahedron']

    """
    print(f"- parameters:  # {index}")
    print(
        "    scattering_approximation:",
        selfen["scattering_approximation"][()].decode("utf-8"),
    )
    print(f"    static_approximation: {bool(selfen['static'][()])}")
    print(f"    use_tetrahedron_method: {bool(selfen["tetrahedron"][()])}")
    if not selfen["tetrahedron"][()]:
        print(f"    smearing_width: {selfen['delta'][()]}")
    print(
        f"    band_start_stop: [{selfen['band_start'][()]}, {selfen['band_stop'][()]}]"
    )
    print(f"    nbands: {selfen['nbands'][()]}")
    print(f"    nbands_sum: {selfen['nbands_sum'][()]}")
    print(f"    nw: {selfen['nw'][()]}")
    print("    temperatures:")
    for i, t in enumerate(selfen["temps"]):
        print(f"    - {t}  # {i + 1}")

    print("  data_array_shapes:")
    print(f"    carrier_per_cell: {list(selfen["carrier_per_cell"].shape)}")
    print(f"    Fan_self_energy: {list(selfen["selfen_fan"].shape)}")
    print(f"    sampling_energy_points: {list(selfen["energies"].shape)}")
    print(f"    Fermi_energies: {list(selfen["efermi"].shape)}")



"""Implementation of velph-selfenergy-generate."""

import pathlib
import shutil
from typing import Literal, Optional

import click
import h5py
import numpy as np
import tomli
from phonopy.interface.calculator import write_crystal_structure
from phonopy.structure.atoms import parse_cell_dict

from phelel.velph.cli.utils import (
    assert_kpoints_mesh_symmetry,
    get_scheduler_dict,
    write_incar,
    write_kpoints_mesh_mode,
    write_launch_script,
)


def write_input_files(toml_filepath: pathlib.Path, dry_run: bool):
    """Generate el-ph input files."""
    write_selfenergy_input_files(toml_filepath, dry_run, calc_type="selfenergy")


def write_selfenergy_input_files(
    toml_filepath: pathlib.Path,
    dry_run: bool,
    calc_type: Optional[Literal["transport", "selfenergy"]] = None,
    current_directory: pathlib.Path = pathlib.Path(""),
    energy_threshold: float = 0.5,
) -> None:
    """Generate el-ph input files.

    current_directory : Path
        Used for test.
    energy_threshold : float
        Energy threshold (gap) to estimate elph_nbands used for
        calc_type="transport".

    """
    directory_path = pathlib.Path(calc_type)

    with open(toml_filepath, "rb") as f:
        toml_dict = tomli.load(f)

    if "vasp" not in toml_dict:
        click.echo(f"[vasp] section not found in {toml_filepath}.", err=True)
        return None

    if "phelel" in toml_dict["vasp"]:
        hdf5_filepath = pathlib.Path("phelel/phelel_params.hdf5")
    elif "supercell" in toml_dict["vasp"]:
        hdf5_filepath = pathlib.Path("supercell/phelel_params.hdf5")
    else:
        click.echo(
            f'[vasp.phelel] section not found in "{toml_filepath}".',
            err=True,
        )
        return None

    if calc_type not in toml_dict["vasp"]:
        click.echo(
            f'[vasp.{calc_type}] section not found in "{toml_filepath}".',
            err=True,
        )
        return None

    # Check phelel_params.hdf5
    if not dry_run and not hdf5_filepath.exists():
        click.echo(f'"{hdf5_filepath}" not found.', err=True)
        click.echo('Run "velph phelel differentiate" if necessary.', err=True)
        return None

    # mkdir, e.g., selfenergy
    directory_path.mkdir(parents=True, exist_ok=True)

    # Dry run
    toml_incar_dict = toml_dict["vasp"][calc_type]["incar"]
    if dry_run:
        toml_incar_dict["nelm"] = 0
        toml_incar_dict["elph_run"] = False

    # Automatic elph_nbands setting for transport.
    if calc_type == "transport":
        # Here toml_incar_dict is updated for setting elph_nbands
        band_index = _find_elph_nbands(current_directory, energy_threshold)
        if band_index is not None:
            click.echo(f'  "elph_nbands={band_index + 1}" in INCAR is set.')
            toml_incar_dict["elph_nbands"] = band_index + 1

    # POSCAR
    primitive = parse_cell_dict(toml_dict["primitive_cell"])
    write_crystal_structure(directory_path / "POSCAR", primitive)

    # INCAR
    write_incar(toml_incar_dict, directory_path, cell=primitive)

    # KPOINTS
    kpoints_dict = toml_dict["vasp"][calc_type]["kpoints"]
    assert_kpoints_mesh_symmetry(toml_dict, kpoints_dict, primitive)
    write_kpoints_mesh_mode(
        toml_incar_dict,
        directory_path,
        f"vasp.{calc_type}.kpoints",
        kpoints_dict,
    )

    # KPOINTS_DENSE
    kpoints_dense_dict = toml_dict["vasp"][calc_type]["kpoints_dense"]
    assert_kpoints_mesh_symmetry(toml_dict, kpoints_dense_dict, primitive)
    write_kpoints_mesh_mode(
        toml_incar_dict,
        directory_path,
        f"vasp.{calc_type}.kpoints_dense",
        kpoints_dense_dict,
        kpoints_filename="KPOINTS_DENSE",
        kspacing_name="elph_kspacing",
    )

    # POTCAR
    potcar_path = pathlib.Path("POTCAR")
    if potcar_path.exists():
        shutil.copy2(potcar_path, directory_path / potcar_path)

    # phelel_params.hdf5
    if not dry_run:
        shutil.copy2(hdf5_filepath, directory_path / hdf5_filepath.name)

    # Scheduler launch script
    if "scheduler" in toml_dict:
        scheduler_dict = get_scheduler_dict(toml_dict, calc_type)
        write_launch_script(scheduler_dict, directory_path)

    click.echo(f'VASP input files were generated in "{directory_path}".')


def _find_elph_nbands(current_directory: pathlib.Path, energy_threshold: float) -> int:
    dos_directory = current_directory / "el_bands" / "dos"
    if dos_directory.exists():
        click.echo('Found "el_bands/dos" directory. Estimate elph_nbands.')
        vaspout_path = dos_directory / "vaspout.h5"
        if vaspout_path.exists():
            possiblly_occupied_band_index = _estimate_elph_nbands(
                vaspout_path, energy_threshold=energy_threshold
            )
            if possiblly_occupied_band_index is None:
                click.echo("Estimation of elph_nbands failed.")
                return None
            return possiblly_occupied_band_index
        else:
            click.echo('Not found "el_bands/dos/vasprun.xml".')
            return None


def _estimate_elph_nbands(
    vaspout_path: pathlib.Path,
    energy_threshold: float = 0.5,
    occupation_condition: float = 1e-5,
    unoccupation_condition: float = 1e-10,
) -> Optional[int]:
    """Estimate elph_nbands from eigenvalues in electronic DOS calculation.

    Parameters
    ----------
    vaspout_path : pathlib.Path
        "vaspout.h5" path.
    energy_threshold : float
        Energy threshold (gap) in eV to estimate elph_nbands. Default is 0.5 eV.
    occupation_condition : float
        Condition to determine either if bands are occupied or not. This value
        is used as occupation_number > occupation_condition. Default is 1e-5.
    unoccupation_condition : float
        Condition to determine either if bands are unoccupied or not. This value
        is used as occupation_number < unoccupation_condition. Default is 1e-10.

    Returns
    -------
    int or None
        Possibly occupied band index by counting from 0. Even if
        occupation number is zero, if some of bands can be occuped by
        excitation, this band index is returned.

    """
    with h5py.File(vaspout_path) as f:
        eigenvalues = f["results"]["electron_eigenvalues_kpoints_opt"]["eigenvalues"][:]
        occupations = f["results"]["electron_eigenvalues_kpoints_opt"]["fermiweights"][
            :
        ]

    nbands = eigenvalues.shape[2]
    unoccupied_band_index = None
    for i in range(nbands):
        if (occupations[:, :, i] < unoccupation_condition).all():
            unoccupied_band_index = i
            break
    if unoccupied_band_index is None:
        return None

    occupied_eigvals = np.sort(
        np.extract(
            occupations[:, :, unoccupied_band_index - 1] > occupation_condition,
            eigenvalues[:, :, unoccupied_band_index - 1],
        )
    )
    max_occupied_eigval = np.max(occupied_eigvals)

    for band_index in range(unoccupied_band_index, nbands):
        min_unoccupied = np.min(eigenvalues[:, :, band_index])
        if max_occupied_eigval + energy_threshold < min_unoccupied:
            return band_index - 1

    return None



"""velph command line tool module."""

import click


@click.group()
@click.help_option("-h", "--help")
def cmd_root():
    """Command-line utility to help VASP el-ph calculation."""
    pass


from phelel.velph.cli.el_bands import cmd_el_bands  # noqa F401
from phelel.velph.cli.generate import cmd_generate  # noqa F401
from phelel.velph.cli.hints import cmd_hints  # noqa F401
from phelel.velph.cli.init import cmd_init  # noqa F401
from phelel.velph.cli.phono3py import cmd_phono3py  # noqa F401
from phelel.velph.cli.nac import cmd_nac  # noqa F401
from phelel.velph.cli.ph_bands import cmd_ph_bands  # noqa F401
from phelel.velph.cli.relax import cmd_relax  # noqa F401
from phelel.velph.cli.selfenergy import cmd_selfenergy  # noqa F401
from phelel.velph.cli.phelel import cmd_phelel  # noqa F401
from phelel.velph.cli.transport import cmd_transport  # noqa F401



"""Implementation of velph-phono3py-init."""

from __future__ import annotations

import pathlib
from typing import Optional, Union

import click
import numpy as np
from phono3py import Phono3py
from phono3py.interface.calculator import get_default_displacement_distance
from phonopy.structure.atoms import parse_cell_dict

from phelel.velph.cli.utils import get_nac_params


def run_init(
    toml_dict: dict,
    current_directory: pathlib.Path = pathlib.Path(""),
    number_of_snapshots: Optional[int] = None,
) -> Optional[Phono3py]:
    """Generate displacements and write phono3py_disp.yaml.

    current_directory : Path
        Used for test.

    """
    convcell = parse_cell_dict(toml_dict["unitcell"])
    supercell_matrix = toml_dict["phelel"].get("supercell_dimension", None)
    if "phonopy" in toml_dict:
        phonon_supercell_matrix = toml_dict["phonopy"].get("supercell_dimension", None)
    else:
        phonon_supercell_matrix = None
    if "primitive_cell" in toml_dict:
        primitive = parse_cell_dict(toml_dict["primitive_cell"])
        primitive_matrix = np.dot(np.linalg.inv(convcell.cell.T), primitive.cell.T)
    else:
        primitive = convcell
        primitive_matrix = None

    is_symmetry = True
    try:
        if toml_dict["phelel"]["nosym"] is True:
            is_symmetry = False
    except KeyError:
        pass

    ph3py = Phono3py(
        convcell,
        supercell_matrix=supercell_matrix,
        phonon_supercell_matrix=phonon_supercell_matrix,
        primitive_matrix=primitive_matrix,
        is_symmetry=is_symmetry,
        calculator="vasp",
    )

    amplitude = toml_dict["phelel"].get("amplitude", None)

    if number_of_snapshots is None:
        is_diagonal = toml_dict["phelel"].get("diagonal", True)
        is_plusminus = toml_dict["phelel"].get("plusminus", "auto")
    else:
        is_diagonal = False
        is_plusminus = False

    _generate_phono3py_supercells(
        ph3py,
        interface_mode="vasp",
        distance=amplitude,
        is_plusminus=is_plusminus,
        is_diagonal=is_diagonal,
        number_of_snapshots=number_of_snapshots,
    )

    nac_directory = current_directory / "nac"
    if nac_directory.exists():
        click.echo('Found "nac" directory. Read NAC params.')
        vasprun_path = nac_directory / "vasprun.xml"
        if vasprun_path.exists():
            nac_params = get_nac_params(
                toml_dict,
                vasprun_path,
                primitive,
                convcell,
                is_symmetry,
            )
            if nac_params is not None:
                ph3py.nac_params = nac_params
        else:
            click.echo('Not found "nac/vasprun.xml". NAC params were not included.')

    return ph3py


def _generate_phono3py_supercells(
    phono3py: Phono3py,
    interface_mode: str = "vasp",
    distance: Optional[float] = None,
    is_plusminus: Union[str, bool] = "auto",
    is_diagonal: bool = True,
    number_of_snapshots: Optional[int] = None,
):
    """Generate phelel supercells."""
    if distance is None:
        _distance = get_default_displacement_distance(interface_mode)
    else:
        _distance = distance

    phono3py.generate_displacements(
        distance=_distance,
        is_plusminus=is_plusminus,
        is_diagonal=is_diagonal,
        number_of_snapshots=number_of_snapshots,
    )
    click.echo(f"Displacement distance: {_distance}")
    click.echo(
        f"Number of displacements: {len(phono3py.supercells_with_displacements)}"
    )

    if phono3py.phonon_supercell_matrix is not None:
        # For estimating number of displacements for harmonic phonon
        phono3py.generate_fc2_displacements(
            distance=distance, is_plusminus=False, is_diagonal=False
        )
        n_disps = len(phono3py.phonon_supercells_with_displacements)
        phono3py.generate_fc2_displacements(
            distance=distance, number_of_snapshots=n_disps, is_plusminus=True
        )
        n_snapshots = len(phono3py.phonon_supercells_with_displacements)
        click.echo(f"Number of displacements for phonon: {n_snapshots}")
